from __future__ import annotations

import importlib
from contextlib import contextmanager
from pathlib import Path

import pytest
from pytest import param

import ibis
from ibis.backends.tests.tpc.conftest import compare_tpc_results
from ibis.formats.pandas import PandasData

tpch_catalog = {
    "lineitem": {
        "l_orderkey": "int32",
        "l_partkey": "int32",
        "l_suppkey": "int32",
        "l_linenumber": "int32",
        "l_quantity": "decimal(15, 2)",
        "l_extendedprice": "decimal(15, 2)",
        "l_discount": "decimal(15, 2)",
        "l_tax": "decimal(15, 2)",
        "l_returnflag": "string",
        "l_linestatus": "string",
        "l_shipdate": "date",
        "l_commitdate": "date",
        "l_receiptdate": "date",
        "l_shipinstruct": "string",
        "l_shipmode": "string",
        "l_comment": "string",
    },
    "customer": [
        ("c_custkey", "int64"),
        ("c_name", "string"),
        ("c_address", "string"),
        ("c_nationkey", "int16"),
        ("c_phone", "string"),
        ("c_acctbal", "decimal"),
        ("c_mktsegment", "string"),
        ("c_comment", "string"),
    ],
    "orders": [
        ("o_orderkey", "int64"),
        ("o_custkey", "int64"),
        ("o_orderstatus", "string"),
        ("o_totalprice", "decimal(12,2)"),
        ("o_orderdate", "date"),
        ("o_orderpriority", "string"),
        ("o_clerk", "string"),
        ("o_shippriority", "int32"),
        ("o_comment", "string"),
    ],
}

root = Path(__file__).absolute().parents[3]

SQL_QUERY_PATH = root / "backends" / "tests" / "tpc" / "queries" / "duckdb" / "h"


@contextmanager
def set_database(con, db):
    olddb = con.current_database
    con.raw_sql(f"USE {db}")
    yield
    con.raw_sql(f"USE {olddb}")


@pytest.mark.parametrize(
    "tpch_query",
    [
        param(1, id="tpch01"),
        param(3, id="tpch03"),
    ],
)
def test_parse_sql_tpch(tpch_query, snapshot, con, data_dir):
    tpch_query_file = SQL_QUERY_PATH / f"{tpch_query:02d}.sql"
    with open(tpch_query_file) as f:
        sql = f.read()

    expr = ibis.parse_sql(sql, tpch_catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "out_tpch.py")

    # Import just-created snapshot
    SNAPSHOT_MODULE = f"ibis.backends.duckdb.tests.snapshots.test_decompile_tpch.test_parse_sql_tpch.tpch{tpch_query:02d}.out_tpch"
    module = importlib.import_module(SNAPSHOT_MODULE)

    with set_database(con, "tpch"):
        # Get results from executing SQL directly on DuckDB
        expected_df = con.con.execute(sql).df()
        # Get results from decompiled ibis query
        result_df = con.to_pandas(module.result)

    # Then set the expected columns so we can coerce the datatypes
    # of the pandas dataframe correctly
    expected_df.columns = result_df.columns

    expected_df = PandasData.convert_table(expected_df, module.result.schema())

    compare_tpc_results(result_df, expected_df)



from __future__ import annotations

import hypothesis as h

import ibis.tests.strategies as its
from ibis.backends.datafusion import as_nullable


def is_nullable(dtype):
    if dtype.is_struct():
        return all(map(is_nullable, dtype.values()))
    elif dtype.is_array():
        return is_nullable(dtype.value_type)
    elif dtype.is_map():
        return is_nullable(dtype.key_type) and is_nullable(dtype.value_type)
    else:
        return dtype.nullable is True


@h.given(its.all_dtypes())
def test_as_nullable(dtype):
    nullable_dtype = as_nullable(dtype)
    assert nullable_dtype.nullable is True
    assert is_nullable(nullable_dtype)



"""Databricks backend."""

from __future__ import annotations

import contextlib
import functools
import getpass
import os
import sys
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any

import databricks.sql
import pyarrow as pa
import pyarrow_hotfix  # noqa: F401
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as exc
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanCreateDatabase, UrlFromPath
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import STAR, AlterTable, C, RenameTable

if TYPE_CHECKING:
    from collections.abc import Mapping

    import pandas as pd
    import polars as pl

    from ibis.expr.schema import SchemaLike


class Backend(SQLBackend, CanCreateDatabase, UrlFromPath):
    name = "databricks"
    compiler = sc.databricks.compiler

    @property
    def current_catalog(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.current_catalog())) as cur:
            [(db,)] = cur.fetchall()
        return db

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.current_database())) as cur:
            [(db,)] = cur.fetchall()
        return db

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)
        cur = self.con.cursor()
        try:
            cur.execute(query, **kwargs)
        except Exception:
            cur.close()
            raise
        return cur

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
        using: str = "delta",
        location: str | None = None,
        tblproperties: Mapping[str, str] | None = None,
    ):
        """Create a table in Databricks.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.

            For multi-level table hierarchies, you can pass in a dotted string
            path like `"catalog.database"` or a tuple of strings like
            `("catalog", "database")`.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists
        using
            Data source format
        location
            Storage location for the table
        tblproperties
            Table properties
        """
        if temp:
            raise exc.UnsupportedOperationError("Temporary tables not yet supported")

        table_loc = self._to_sqlglot_table(database)

        catalog = table_loc.catalog or self.current_catalog
        database = table_loc.db or self.current_database

        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        properties = [sge.FileFormatProperty(this=self.compiler.v[using.upper()])]

        if location is not None:
            properties.append(sge.LocationProperty(this=sge.convert(location)))

        for key, value in (tblproperties or {}).items():
            properties.append(
                sge.Property(this=sge.convert(str(key)), value=sge.convert(str(value)))
            )

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name("databricks_table")
        else:
            temp_name = name

        quoted = self.compiler.quoted
        dialect = self.dialect

        initial_table = sg.table(temp_name, catalog=catalog, db=database, quoted=quoted)
        target = sge.Schema(
            this=initial_table,
            expressions=(schema or table.schema()).to_sqlglot(dialect),
        )

        properties = sge.Properties(expressions=properties)
        create_stmt = sge.Create(kind="TABLE", this=target, properties=properties)

        # This is the same table as initial_table unless overwrite == True
        final_table = sg.table(name, catalog=catalog, db=database, quoted=quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                insert_stmt = sge.insert(query, into=initial_table).sql(dialect)
                cur.execute(insert_stmt).fetchall()

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=final_table, exists=True).sql(dialect)
                )
                if temp:
                    cur.execute(
                        sge.Create(
                            kind="TABLE",
                            this=final_table,
                            expression=sg.select(STAR).from_(initial_table),
                            properties=properties,
                        ).sql(dialect)
                    )
                    cur.execute(
                        sge.Drop(kind="TABLE", this=initial_table, exists=True).sql(
                            dialect
                        )
                    )
                else:
                    cur.execute(
                        AlterTable(
                            this=initial_table, actions=[RenameTable(this=final_table)]
                        ).sql(dialect)
                    )

        return self.table(name, database=(catalog, database))

    def table(self, name: str, database: str | None = None) -> ir.Table:
        """Construct a table expression.

        Parameters
        ----------
        name
            Table name
        database
            Database name

        Returns
        -------
        Table
            Table expression

        """
        table_loc = self._to_sqlglot_table(database)

        # TODO: set these to better defaults
        catalog = table_loc.catalog or None
        database = table_loc.db or None

        table_schema = self.get_schema(name, catalog=catalog, database=database)
        return ops.DatabaseTable(
            name,
            schema=table_schema,
            source=self,
            namespace=ops.Namespace(catalog=catalog, database=database),
        ).to_expr()

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        """Compute the schema of a `table`.

        Parameters
        ----------
        table_name
            May **not** be fully qualified. Use `database` if you want to
            qualify the identifier.
        catalog
            Catalog name
        database
            Database name

        Returns
        -------
        sch.Schema
            Ibis schema
        """
        table = sg.table(
            table_name, db=database, catalog=catalog, quoted=self.compiler.quoted
        )
        sql = sge.Describe(kind="TABLE", this=table).sql(self.dialect)
        try:
            with self.con.cursor() as cur:
                out = cur.execute(sql).fetchall_arrow()
        except databricks.sql.exc.ServerOperationError as e:
            raise exc.TableNotFound(
                f"Table {table_name!r} not found in "
                f"{catalog or self.current_catalog}.{database or self.current_database}"
            ) from e

        names = out["col_name"].to_pylist()
        types = out["data_type"].to_pylist()

        return sch.Schema(
            dict(zip(names, map(self.compiler.type_mapper.from_string, types)))
        )

    @contextlib.contextmanager
    def _safe_raw_sql(self, query, *args, **kwargs):
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)
        with self.con.cursor() as cur:
            yield cur.execute(query, *args, **kwargs)

    def list_catalogs(self, like: str | None = None) -> list[str]:
        with self.con.cursor() as cur:
            out = cur.catalogs().fetchall_arrow()
        return self._filter_with_like(out["TABLE_CAT"].to_pylist(), like)

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        with self.con.cursor() as cur:
            out = cur.schemas(
                catalog_name=catalog or self.current_catalog
            ).fetchall_arrow()
        return self._filter_with_like(out["TABLE_SCHEM"].to_pylist(), like=like)

    @functools.cached_property
    def version(self) -> str:
        query = sg.select(self.compiler.f.current_version())
        with self._safe_raw_sql(query) as cur:
            [(version_info,)] = cur.fetchall()
        return version_info["dbsql_version"]

    def do_connect(
        self,
        *,
        server_hostname: str | None = None,
        http_path: str | None = None,
        access_token: str | None = None,
        auth_type: str | None = None,
        credentials_provider: str | None = None,
        password: str | None = None,
        username: str | None = None,
        session_configuration: Mapping[str, str] | None = None,
        http_headers: list[tuple[str, str]] | None = None,
        catalog: str | None = None,
        schema: str = "default",
        use_cloud_fetch: bool = False,
        memtable_volume: str | None = None,
        staging_allowed_local_path: str | None = None,
        **config: Any,
    ) -> None:
        """Create an Ibis client connected to a Databricks cloud instance."""
        if staging_allowed_local_path is None:
            staging_allowed_local_path = tempfile.gettempdir()
        self.con = databricks.sql.connect(
            server_hostname=(
                server_hostname or os.environ.get("DATABRICKS_SERVER_HOSTNAME")
            ),
            http_path=http_path or os.environ.get("DATABRICKS_HTTP_PATH"),
            access_token=access_token or os.environ.get("DATABRICKS_TOKEN"),
            auth_type=auth_type,
            credentials_provider=credentials_provider,
            password=password,
            username=username,
            session_configuration=session_configuration,
            http_headers=http_headers,
            catalog=catalog,
            schema=schema,
            use_cloud_fetch=use_cloud_fetch,
            staging_allowed_local_path=staging_allowed_local_path,
            **config,
        )
        if memtable_volume is None:
            short_version = "".join(map(str, sys.version_info[:3]))
            memtable_volume = (
                f"{getpass.getuser()}-py={short_version}-pid={os.getpid()}"
            )
        self._memtable_volume = memtable_volume
        self._memtable_catalog = self.current_catalog
        self._memtable_database = self.current_database
        self._post_connect(memtable_volume=memtable_volume)

    @contextlib.contextmanager
    def begin(self):
        with self.con.cursor() as cur:
            yield cur

    @util.experimental
    @classmethod
    def from_connection(
        cls,
        con,
        memtable_volume: str | None = None,
    ) -> Backend:
        """Create an Ibis client from an existing connection to a Databricks cloud instance.

        Parameters
        ----------
        con
            An existing connection to a Databricks database.
        memtable_volume
            The volume to use for Ibis memtables.
        """
        new_backend = cls()
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect(memtable_volume=memtable_volume)
        return new_backend

    def _post_connect(self, *, memtable_volume: str) -> None:
        sql = f"CREATE VOLUME IF NOT EXISTS `{memtable_volume}` COMMENT 'Ibis memtable storage volume'"
        with self.con.cursor() as cur:
            cur.execute(sql)

    @functools.cached_property
    def _memtable_volume_path(self) -> str:
        return f"/Volumes/{self._memtable_catalog}/{self._memtable_database}/{self._memtable_volume}"

    def _in_memory_table_exists(self, name: str) -> bool:
        sql = (
            sg.select(self.compiler.f.count(STAR))
            .from_(
                sg.table("views", db="information_schema", catalog=self.current_catalog)
            )
            .where(
                C.table_name.eq(sge.convert(name)),
                C.table_schema.eq(self.compiler.f.current_database()),
            )
        )
        with self._safe_raw_sql(sql) as cur:
            [(out,)] = cur.fetchall()

        assert 0 <= out <= 1, str(out)
        return out == 1

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        import pyarrow.parquet as pq

        quoted = self.compiler.quoted
        name = op.name
        stem = f"{name}.parquet"

        upstream_path = f"{self._memtable_volume_path}/{stem}"
        sql = sge.Create(
            kind="VIEW",
            this=sg.table(
                name,
                db=self.current_database,
                catalog=self.current_catalog,
                quoted=quoted,
            ),
            expression=sge.select(STAR).from_(
                sg.table(upstream_path, db="parquet", quoted=quoted)
            ),
        ).sql(self.dialect)
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            data = op.data.to_pyarrow(schema=op.schema)
            path = Path(tmpdir, stem)
            put_into = f"PUT '{path}' INTO '{upstream_path}' OVERWRITE"
            # optimize for bandwidth so use zstd which typically compresses
            # better than the other options without much loss in speed
            pq.write_table(data, path, compression="zstd")
            with self.con.cursor() as cur:
                cur.execute(put_into)
                cur.execute(sql)

    def _finalize_memtable(self, name: str) -> None:
        path = f"{self._memtable_volume_path}/{name}.parquet"
        sql = sge.Drop(
            kind="VIEW",
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            exists=True,
        ).sql(self.dialect)
        with self.con.cursor() as cur:
            cur.execute(sql)
            cur.execute(f"REMOVE '{path}'")

    def create_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        name = sg.table(name, catalog=catalog, quoted=self.compiler.quoted)
        with self._safe_raw_sql(sge.Create(this=name, kind="SCHEMA", replace=force)):
            pass

    def drop_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        name = sg.table(name, catalog=catalog, quoted=self.compiler.quoted)
        with self._safe_raw_sql(sge.Drop(this=name, kind="SCHEMA", replace=force)):
            pass

    def list_tables(
        self,
        like: str | None = None,
        database: tuple[str, str] | str | None = None,
    ) -> list[str]:
        """List tables and views.

        ::: {.callout-note}
        ## Ibis does not use the word `schema` to refer to database hierarchy.

        A collection of tables is referred to as a `database`.
        A collection of `database` is referred to as a `catalog`.

        These terms are mapped onto the corresponding features in each
        backend (where available), regardless of whether the backend itself
        uses the same terminology.
        :::

        Parameters
        ----------
        like
            Regex to filter by table/view name.
        database
            Database location. If not passed, uses the current database.

            By default uses the current `database` (`self.current_database`) and
            `catalog` (`self.current_catalog`).

            To specify a table in a separate catalog, you can pass in the
            catalog and database as a string `"catalog.database"`, or as a tuple of
            strings `("catalog", "database")`.

        Returns
        -------
        list[str]
            List of table and view names.

        Examples
        --------
        >>> import ibis
        >>> con = ibis.databricks.connect()
        >>> foo = con.create_table("foo", schema=ibis.schema(dict(a="int")))
        >>> con.list_tables()
        ['foo']
        >>> bar = con.create_view("bar", foo)
        >>> con.list_tables()
        ['bar', 'foo']
        >>> con.create_database("my_database")
        >>> con.list_tables(database="my_database")
        []
        >>> con.raw_sql("CREATE TABLE my_database.baz (a INTEGER)")  # doctest: +ELLIPSIS
        <... object at 0x...>
        >>> con.list_tables(database="my_database")
        ['baz']

        """
        table_loc = self._to_sqlglot_table(database)

        catalog = table_loc.catalog or self.current_catalog
        database = table_loc.db or self.current_database

        with self.con.cursor() as cur:
            cur.tables(catalog_name=catalog, schema_name=database)
            out = cur.fetchall_arrow()

        return self._filter_with_like(out["TABLE_NAME"].to_pylist(), like)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        """Return a stream of record batches.

        The returned `RecordBatchReader` contains a cursor with an unbounded lifetime.

        For analytics use cases this is usually nothing to fret about. In some cases you
        may need to explicit release the cursor.

        Parameters
        ----------
        expr
            Ibis expression
        params
            Bound parameters
        limit
            Limit the result to this number of rows
        chunk_size
            The number of rows to fetch per batch
        """
        self._run_pre_execute_hooks(expr)
        table = expr.as_table()
        sql = self.compile(table, limit=limit, params=params)

        def batch_producer(con, sql):
            with con.cursor() as cur:
                batched_cur = cur.execute(sql)
                while batch := batched_cur.fetchmany_arrow(size=chunk_size):
                    yield from batch.to_batches()

        pyarrow_schema = expr.as_table().schema().to_pyarrow()
        producer = batch_producer(self.con, sql)
        return pa.ipc.RecordBatchReader.from_batches(pyarrow_schema, producer)

    def to_pyarrow(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        self._run_pre_execute_hooks(expr)

        sql = self.compile(expr, limit=limit, params=params, **kwargs)
        with self._safe_raw_sql(sql) as cur:
            res = cur.fetchall_arrow()

        target_schema = expr.as_table().schema().to_pyarrow()
        if res is None:
            res = target_schema.empty_table()

        return expr.__pyarrow_result__(res)

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        if (table := cursor.fetchall_arrow()) is None:
            table = schema.to_pyarrow().empty_table()
        df = table.to_pandas(timestamp_as_object=True)
        df.columns = list(schema.names)
        return df

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        with self._safe_raw_sql(
            sge.Describe(this=sg.parse_one(query, read=self.dialect))
        ) as cur:
            rows = cur.fetchall_arrow()

        rows = rows.to_pydict()

        type_mapper = self.compiler.type_mapper
        return sch.Schema(
            {
                name: type_mapper.from_string(typ, nullable=True)
                for name, typ in zip(rows["col_name"], rows["data_type"])
            }
        )

    def _get_temp_view_definition(self, name: str, definition: str) -> str:
        return sge.Create(
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            kind="VIEW",
            expression=definition,
            replace=True,
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        )

    def _create_temp_view(self, table_name, source):
        with self._safe_raw_sql(self._get_temp_view_definition(table_name, source)):
            pass

    def rename_table(self, old_name: str, new_name: str) -> None:
        """Rename an existing table.

        Parameters
        ----------
        old_name
            The old name of the table.
        new_name
            The new name of the table.

        """
        old = sg.table(old_name, quoted=True)
        new = sg.table(new_name, quoted=True)
        query = AlterTable(
            this=old, exists=False, actions=[RenameTable(this=new, exists=True)]
        )
        with self._safe_raw_sql(query):
            pass



from __future__ import annotations

import itertools
from datetime import date
from operator import methodcaller

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
from ibis import _
from ibis import literal as L
from ibis.backends.tests.errors import (
    ClickHouseDatabaseError,
    DatabricksServerOperationError,
    ExaQueryError,
    GoogleBadRequest,
    ImpalaHiveServer2Error,
    MySQLNotSupportedError,
    OracleDatabaseError,
    PolarsInvalidOperationError,
    PsycoPg2InternalError,
    Py4JError,
    Py4JJavaError,
    PyDruidProgrammingError,
    PyODBCProgrammingError,
    PySparkAnalysisException,
    PySparkPythonException,
    SnowflakeProgrammingError,
    TrinoUserError,
)
from ibis.conftest import IS_SPARK_REMOTE
from ibis.legacy.udf.vectorized import reduction

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")

with pytest.warns(FutureWarning, match="v9.0"):

    @reduction(input_type=[dt.double], output_type=dt.double)
    def mean_udf(s):
        return s.mean()


aggregate_test_params = [
    param(lambda t: t.double_col.mean(), lambda t: t.double_col.mean(), id="mean"),
    param(
        lambda t: mean_udf(t.double_col),
        lambda t: t.double_col.mean(),
        id="mean_udf",
        marks=[
            pytest.mark.notimpl(
                [
                    "bigquery",
                    "datafusion",
                    "postgres",
                    "risingwave",
                    "clickhouse",
                    "impala",
                    "duckdb",
                    "polars",
                    "snowflake",
                    "mssql",
                    "trino",
                    "druid",
                    "oracle",
                    "flink",
                    "exasol",
                    "databricks",
                ],
                raises=com.OperationNotDefinedError,
            ),
            pytest.mark.never(
                ["sqlite", "mysql"],
                reason="no udf support",
                raises=com.OperationNotDefinedError,
            ),
            pytest.mark.notyet(
                ["pyspark"],
                condition=IS_SPARK_REMOTE,
                raises=PySparkPythonException,
                reason="remote udfs not yet tested due to environment complexities",
            ),
        ],
    ),
    param(lambda t: t.double_col.min(), lambda t: t.double_col.min(), id="min"),
    param(lambda t: t.double_col.max(), lambda t: t.double_col.max(), id="max"),
    param(
        # int_col % 3 so there are no ties for most common value
        lambda t: (t.int_col % 3).mode(),
        lambda t: (t.int_col % 3).mode().iloc[0],
        id="mode",
        marks=[
            pytest.mark.notyet(
                [
                    "bigquery",
                    "clickhouse",
                    "datafusion",
                    "impala",
                    "mysql",
                    "mssql",
                    "pyspark",
                    "trino",
                    "druid",
                    "flink",
                    "risingwave",
                    "exasol",
                ],
                raises=com.OperationNotDefinedError,
            ),
        ],
    ),
    param(
        lambda t: (t.double_col + 5).sum(),
        lambda t: (t.double_col + 5).sum(),
        id="complex_sum",
    ),
    param(
        lambda t: t.timestamp_col.max(),
        lambda t: t.timestamp_col.max(),
        id="timestamp_max",
    ),
]

argidx_not_grouped_marks = [
    "impala",
    "mysql",
    "mssql",
    "druid",
    "oracle",
    "flink",
    "exasol",
]


def make_argidx_params(marks):
    marks = [pytest.mark.notyet(marks, raises=com.OperationNotDefinedError)]
    return [
        param(
            lambda t: t.timestamp_col.argmin(t.id),
            lambda s: s.timestamp_col.iloc[s.id.argmin()],
            id="argmin",
            marks=marks,
        ),
        param(
            lambda t: t.double_col.argmax(t.id),
            lambda s: s.double_col.iloc[s.id.argmax()],
            id="argmax",
            marks=marks,
        ),
    ]


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    aggregate_test_params + make_argidx_params(argidx_not_grouped_marks),
)
def test_aggregate(backend, alltypes, df, result_fn, expected_fn):
    expr = alltypes.aggregate(tmp=result_fn)
    result = expr.execute()

    # Create a single-row single-column dataframe with the Pandas `agg` result
    # (to match the output format of Ibis `aggregate`)
    expected = pd.DataFrame({"tmp": [expected_fn(df)]})

    backend.assert_frame_equal(result, expected, check_dtype=False)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    aggregate_test_params + make_argidx_params(argidx_not_grouped_marks),
)
def test_aggregate_grouped(backend, alltypes, df, result_fn, expected_fn):
    grouping_key_col = "bigint_col"

    # Two (equivalent) variations:
    #  1) `group_by` then `aggregate`
    #  2) `aggregate` with `by`
    expr1 = alltypes.group_by(grouping_key_col).aggregate(tmp=result_fn)
    expr2 = alltypes.aggregate(tmp=result_fn, by=grouping_key_col)
    result1 = expr1.execute()
    result2 = expr2.execute()

    # Note: Using `reset_index` to get the grouping key as a column
    expected = (
        df.groupby(grouping_key_col).apply(expected_fn).rename("tmp").reset_index()
    )

    # Row ordering may differ depending on backend, so sort on the
    # grouping key
    result1 = result1.sort_values(by=grouping_key_col).reset_index(drop=True)
    result2 = result2.sort_values(by=grouping_key_col).reset_index(drop=True)
    expected = expected.sort_values(by=grouping_key_col).reset_index(drop=True)

    backend.assert_frame_equal(result1, expected, check_dtype=False)
    backend.assert_frame_equal(result2, expected, check_dtype=False)


@pytest.mark.notimpl(
    [
        "bigquery",
        "clickhouse",
        "datafusion",
        "duckdb",
        "impala",
        "mysql",
        "postgres",
        "risingwave",
        "sqlite",
        "snowflake",
        "polars",
        "mssql",
        "trino",
        "druid",
        "oracle",
        "flink",
        "exasol",
        "databricks",
    ],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["pyspark"],
    raises=(PySparkPythonException, NotImplementedError),
)
def test_aggregate_multikey_group_reduction_udf(backend, alltypes, df):
    """Tests .aggregate() on a multi-key group_by with a reduction
    operation."""
    with pytest.warns(FutureWarning, match="v9.0"):

        @reduction(
            input_type=[dt.double],
            output_type=dt.Struct({"mean": dt.double, "std": dt.double}),
        )
        def mean_and_std(v):
            return v.mean(), v.std()

    grouping_key_cols = ["bigint_col", "int_col"]

    expr1 = alltypes.group_by(grouping_key_cols).aggregate(
        mean_and_std(alltypes["double_col"]).destructure()
    )

    result1 = expr1.execute()

    # Note: Using `reset_index` to get the grouping key as a column
    expected = (
        df.groupby(grouping_key_cols)["double_col"].agg(["mean", "std"]).reset_index()
    )

    # Row ordering may differ depending on backend, so sort on the
    # grouping key
    result1 = result1.sort_values(by=grouping_key_cols).reset_index(drop=True)
    expected = (
        expected.sort_values(by=grouping_key_cols)
        .reset_index(drop=True)
        .assign(int_col=lambda df: df.int_col.astype("int32"))
    )

    backend.assert_frame_equal(result1, expected)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t, where: t.bool_col.count(where=where),
            lambda t, where: len(t.bool_col[where].dropna()),
            id="count",
        ),
        param(
            lambda t, where: t.bool_col.nunique(where=where),
            lambda t, where: t.bool_col[where].dropna().nunique(),
            id="nunique",
        ),
        param(
            lambda t, where: t.bool_col.any(where=where),
            lambda t, where: t.bool_col[where].any(),
            id="any",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=AttributeError,
                    reason="'IntegerColumn' object has no attribute 'any'",
                ),
            ],
        ),
        param(
            lambda t, where: t.bool_col.notany(where=where),
            lambda t, where: ~t.bool_col[where].any(),
            id="notany",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=AttributeError,
                    reason="'IntegerColumn' object has no attribute 'notany'",
                ),
            ],
        ),
        param(
            lambda t, where: ~t.bool_col.any(where=where),
            lambda t, where: ~t.bool_col[where].any(),
            id="any_negate",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=AttributeError,
                    reason="'IntegerColumn' object has no attribute 'any'",
                ),
            ],
        ),
        param(
            lambda t, where: t.bool_col.all(where=where),
            lambda t, where: t.bool_col[where].all(),
            id="all",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=AttributeError,
                    reason="'IntegerColumn' object has no attribute 'all'",
                ),
            ],
        ),
        param(
            lambda t, where: t.bool_col.notall(where=where),
            lambda t, where: ~t.bool_col[where].all(),
            id="notall",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=AttributeError,
                    reason="'IntegerColumn' object has no attribute 'notall'",
                ),
            ],
        ),
        param(
            lambda t, where: ~t.bool_col.all(where=where),
            lambda t, where: ~t.bool_col[where].all(),
            id="all_negate",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=AttributeError,
                    reason="'IntegerColumn' object has no attribute 'all'",
                ),
            ],
        ),
        param(
            lambda t, where: t.double_col.sum(where=where),
            lambda t, where: t.double_col[where].sum(),
            id="sum",
        ),
        param(
            lambda t, where: (t.int_col > 0).sum(where=where),
            lambda t, where: (t.int_col > 0)[where].sum(),
            id="bool_sum",
            marks=[
                pytest.mark.notimpl(
                    ["oracle"],
                    raises=OracleDatabaseError,
                    reason="ORA-02000: missing AS keyword",
                ),
            ],
        ),
        param(
            lambda t, where: t.double_col.mean(where=where),
            lambda t, where: t.double_col[where].mean(),
            id="mean",
        ),
        param(
            lambda t, where: t.double_col.min(where=where),
            lambda t, where: t.double_col[where].min(),
            id="min",
        ),
        param(
            lambda t, where: t.double_col.max(where=where),
            lambda t, where: t.double_col[where].max(),
            id="max",
        ),
        param(
            # int_col % 3 so there are no ties for most common value
            lambda t, where: (t.int_col % 3).mode(where=where),
            lambda t, where: (t.int_col % 3)[where].mode().iloc[0],
            id="mode",
            marks=[
                pytest.mark.notyet(
                    [
                        "bigquery",
                        "clickhouse",
                        "datafusion",
                        "impala",
                        "mysql",
                        "pyspark",
                        "mssql",
                        "trino",
                        "druid",
                        "exasol",
                        "flink",
                        "risingwave",
                    ],
                    raises=com.OperationNotDefinedError,
                ),
            ],
        ),
        param(
            lambda t, where: t.double_col.argmin(t.int_col, where=where),
            lambda t, where: t.double_col[where].iloc[t.int_col[where].argmin()],
            id="argmin",
            marks=[
                pytest.mark.notyet(
                    [
                        "impala",
                        "mysql",
                        "mssql",
                        "druid",
                        "oracle",
                        "exasol",
                        "flink",
                    ],
                    raises=com.OperationNotDefinedError,
                ),
            ],
        ),
        param(
            lambda t, where: t.double_col.argmax(t.int_col, where=where),
            lambda t, where: t.double_col[where].iloc[t.int_col[where].argmax()],
            id="argmax",
            marks=[
                pytest.mark.notyet(
                    [
                        "impala",
                        "mysql",
                        "mssql",
                        "druid",
                        "oracle",
                        "exasol",
                        "flink",
                    ],
                    raises=com.OperationNotDefinedError,
                ),
            ],
        ),
        param(
            lambda t, where: t.double_col.std(how="sample", where=where),
            lambda t, where: t.double_col[where].std(ddof=1),
            id="std",
            marks=[pytest.mark.notimpl(["druid"], raises=com.OperationNotDefinedError)],
        ),
        param(
            lambda t, where: t.double_col.var(how="sample", where=where),
            lambda t, where: t.double_col[where].var(ddof=1),
            id="var",
            marks=[pytest.mark.notimpl(["druid"], raises=com.OperationNotDefinedError)],
        ),
        param(
            lambda t, where: t.double_col.std(how="pop", where=where),
            lambda t, where: t.double_col[where].std(ddof=0),
            id="std_pop",
            marks=[pytest.mark.notimpl(["druid"], raises=com.OperationNotDefinedError)],
        ),
        param(
            lambda t, where: t.double_col.var(how="pop", where=where),
            lambda t, where: t.double_col[where].var(ddof=0),
            id="var_pop",
            marks=[pytest.mark.notimpl(["druid"], raises=com.OperationNotDefinedError)],
        ),
        param(
            lambda t, where: t.string_col.approx_nunique(where=where),
            lambda t, where: t.string_col[where].nunique(),
            id="approx_nunique",
            marks=[
                pytest.mark.xfail_version(
                    duckdb=["duckdb>=1.1"],
                    raises=AssertionError,
                    reason="not exact, even at this tiny scale",
                    strict=False,
                ),
                pytest.mark.notimpl(
                    ["datafusion"],
                    reason="data type is not supported",
                    raises=Exception,
                ),
            ],
        ),
        param(
            lambda t, where: t.bigint_col.bit_and(where=where),
            lambda t, where: np.bitwise_and.reduce(t.bigint_col[where].values),
            id="bit_and",
            marks=[
                pytest.mark.notimpl(
                    ["polars", "mssql", "exasol"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notimpl(["druid"], strict=False, raises=AssertionError),
                pytest.mark.notyet(
                    ["impala", "pyspark", "flink"], raises=com.OperationNotDefinedError
                ),
            ],
        ),
        param(
            lambda t, where: t.bigint_col.bit_or(where=where),
            lambda t, where: np.bitwise_or.reduce(t.bigint_col[where].values),
            id="bit_or",
            marks=[
                pytest.mark.notimpl(
                    ["polars", "mssql", "exasol"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["impala", "pyspark", "flink"], raises=com.OperationNotDefinedError
                ),
            ],
        ),
        param(
            lambda t, where: t.bigint_col.bit_xor(where=where),
            lambda t, where: np.bitwise_xor.reduce(t.bigint_col[where].values),
            id="bit_xor",
            marks=[
                pytest.mark.notimpl(
                    ["polars", "mssql", "exasol"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["impala", "pyspark", "flink"], raises=com.OperationNotDefinedError
                ),
            ],
        ),
        param(
            lambda t, where: t.count(where=where),
            lambda t, where: len(t[where]),
            id="count_star",
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
    ],
)
def test_reduction_ops(
    backend,
    alltypes,
    df,
    result_fn,
    expected_fn,
    ibis_cond,
    pandas_cond,
):
    # Operate on a subset of the data, since aggregations like var/std with
    # sample/population can be too numerically similar for a larger number of
    # rows.
    alltypes = alltypes.filter(alltypes.id < 1550)
    df = df[df.id < 1550]

    expr = alltypes.agg(tmp=result_fn(alltypes, ibis_cond(alltypes))).tmp
    result = expr.execute().squeeze()
    expected = expected_fn(df, pandas_cond(df))

    try:
        np.testing.assert_allclose(result, expected, rtol=backend.reduction_tolerance)
    except TypeError:  # assert_allclose only handles numerics
        # if we're not testing numerics, then the arrays should be exactly equal
        np.testing.assert_array_equal(result, expected)


@pytest.mark.notimpl(
    ["druid", "impala", "mssql", "mysql", "oracle"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notimpl(
    ["risingwave"],
    raises=com.UnsupportedOperationError,
    reason="risingwave requires an `order_by` for these aggregations",
)
@pytest.mark.parametrize("method", ["first", "last"])
@pytest.mark.parametrize("filtered", [False, True])
@pytest.mark.parametrize(
    "include_null",
    [
        False,
        param(
            True,
            marks=[
                pytest.mark.notimpl(
                    [
                        "clickhouse",
                        "exasol",
                        "flink",
                        "postgres",
                        "risingwave",
                        "snowflake",
                    ],
                    raises=com.UnsupportedOperationError,
                    reason="`include_null=True` is not supported",
                ),
                pytest.mark.notimpl(
                    ["bigquery", "pyspark", "databricks"],
                    raises=com.UnsupportedOperationError,
                    reason="Can't mix `where` and `include_null=True`",
                    strict=False,
                ),
            ],
        ),
    ],
)
def test_first_last(alltypes, method, filtered, include_null):
    # `first` and `last` effectively choose an arbitrary value when no
    # additional order is specified. *Most* backends will result in the
    # first/last element in a column being selected (at least when operating on
    # a leaf table), but that's really not guaranteed. These operations need an
    # order to be meaningful.
    #
    # To sanely test this we create a column that is a mix of nulls and a
    # single value (or a single value after filtering is applied).
    if filtered:
        new = alltypes.int_col.cases((3, 30), (4, 40))
        where = _.int_col == 3
    else:
        new = (alltypes.int_col == 3).ifelse(30, None)
        where = None

    t = alltypes.mutate(new=new)

    expr = getattr(t.new, method)(where=where, include_null=include_null)
    res = expr.execute()
    if include_null:
        # no ordering, so technically could be either 30 or NULL
        assert res == 30 or pd.isna(res)
    else:
        assert res == 30


@pytest.mark.notimpl(
    ["clickhouse", "exasol", "flink", "pyspark", "sqlite", "databricks"],
    raises=com.UnsupportedOperationError,
)
@pytest.mark.notimpl(
    ["druid", "impala", "mssql", "mysql", "oracle"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.parametrize("method", ["first", "last"])
@pytest.mark.parametrize("filtered", [False, True], ids=["not-filtered", "filtered"])
@pytest.mark.parametrize(
    "include_null",
    [
        param(False, id="exclude-null"),
        param(
            True,
            marks=[
                pytest.mark.notimpl(
                    ["clickhouse", "exasol", "flink", "postgres", "snowflake"],
                    raises=com.UnsupportedOperationError,
                    reason="`include_null=True` is not supported",
                ),
                pytest.mark.notimpl(
                    ["bigquery"],
                    raises=com.UnsupportedOperationError,
                    reason="Can't mix `where` and `include_null=True`",
                    strict=False,
                ),
            ],
            id="include-null",
        ),
    ],
)
def test_first_last_ordered(alltypes, method, filtered, include_null):
    t = alltypes.mutate(new=alltypes.int_col.nullif(0).nullif(9))
    if filtered:
        where = _.int_col != (1 if method == "last" else 8)
        sol = 2 if method == "last" else 7
    else:
        where = None
        sol = 1 if method == "last" else 8

    expr = getattr(t.new, method)(
        where=where, order_by=t.int_col.desc(), include_null=include_null
    )
    res = expr.execute()
    if include_null:
        assert pd.isna(res)
    else:
        assert res == sol


@pytest.mark.notimpl(
    [
        "druid",
        "exasol",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "oracle",
    ],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.parametrize("method", ["argmin", "argmax"])
@pytest.mark.parametrize("filtered", [True, False], ids=["filtered", "unfiltered"])
@pytest.mark.parametrize("null_result", [True, False], ids=["null", "non-null"])
def test_argmin_argmax(alltypes, method, filtered, null_result):
    t = alltypes.mutate(by_col=_.int_col.nullif(0).nullif(9), val_col=10 * _.int_col)

    if filtered:
        where = _.int_col != (1 if method == "argmin" else 8)
        sol = 20 if method == "argmin" else 70
    else:
        where = None
        sol = 10 if method == "argmin" else 80

    if null_result:
        t = t.mutate(val_col=_.val_col.nullif(sol))

    expr = getattr(t.val_col, method)("by_col", where=where)
    res = expr.execute()
    assert pd.isna(res) if null_result else res == sol


@pytest.mark.notimpl(
    [
        "impala",
        "mysql",
        "mssql",
        "druid",
        "oracle",
        "exasol",
        "flink",
        "risingwave",
    ],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.parametrize("filtered", [False, True])
def test_arbitrary(alltypes, filtered):
    # Arbitrary chooses a non-null arbitrary value. To ensure we can test for
    # _something_ we create a column that is a mix of nulls and a single value
    # (or a single value after filtering is applied).
    if filtered:
        new = alltypes.int_col.cases((3, 30), (4, 40))
        where = _.int_col == 3
    else:
        new = (alltypes.int_col == 3).ifelse(30, None)
        where = None

    t = alltypes.mutate(new=new)

    expr = t.new.arbitrary(where=where)
    res = expr.execute()
    assert res == 30


@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="cond",
            marks=[
                pytest.mark.notyet(
                    ["mysql"],
                    raises=com.UnsupportedOperationError,
                    reason="backend does not support filtered count distinct with more than one column",
                ),
            ],
        ),
    ],
)
@pytest.mark.notyet(
    ["druid", "mssql", "oracle", "sqlite", "flink"],
    raises=(
        OracleDatabaseError,
        com.UnsupportedOperationError,
        com.OperationNotDefinedError,
    ),
    reason="backend doesn't support count distinct with multiple columns",
)
@pytest.mark.notyet(
    ["datafusion"],
    raises=com.OperationNotDefinedError,
    reason="no one has attempted implementation yet",
)
def test_count_distinct_star(alltypes, df, ibis_cond, pandas_cond):
    table = alltypes[["int_col", "double_col", "string_col"]]
    expr = table.nunique(where=ibis_cond(table))
    result = expr.execute()
    df = df[["int_col", "double_col", "string_col"]]
    expected = len(df.loc[pandas_cond(df)].drop_duplicates())
    assert int(result) == int(expected)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t, where: t.double_col.quantile(0.5, where=where),
            lambda t, where: t.double_col[where].quantile(0.5),
            id="quantile",
            marks=[
                pytest.mark.notimpl(
                    [
                        "bigquery",
                        "datafusion",
                        "impala",
                        "mssql",
                        "mysql",
                        "sqlite",
                        "druid",
                    ],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.never(
                    ["trino"],
                    reason="backend implements approximate quantiles",
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.never(
                    ["flink"],
                    reason="backend doesn't implement approximate quantiles yet",
                    raises=com.OperationNotDefinedError,
                ),
            ],
        ),
        param(
            lambda t, where: t.double_col.quantile([0.5], where=where),
            lambda t, where: t.double_col[where].quantile([0.5]),
            id="multi-quantile",
            marks=[
                pytest.mark.notimpl(
                    [
                        "bigquery",
                        "datafusion",
                        "polars",
                        "druid",
                        "oracle",
                    ],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["mysql", "mssql", "impala", "exasol", "sqlite"],
                    raises=com.UnsupportedBackendType,
                ),
                pytest.mark.notyet(
                    ["snowflake", "risingwave"],
                    reason="backend doesn't implement array of quantiles as input",
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.never(
                    ["trino"],
                    reason="backend implements approximate quantiles",
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.never(
                    ["flink"],
                    reason="backend doesn't implement approximate quantiles yet",
                    raises=com.OperationNotDefinedError,
                ),
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
            marks=[
                pytest.mark.notimpl(["datafusion"], raises=com.OperationNotDefinedError)
            ],
        ),
    ],
)
def test_quantile(
    alltypes,
    df,
    result_fn,
    expected_fn,
    ibis_cond,
    pandas_cond,
):
    expr = alltypes.agg(tmp=result_fn(alltypes, ibis_cond(alltypes))).tmp
    result = expr.execute().squeeze()
    expected = expected_fn(df, pandas_cond(df))
    assert pytest.approx(result) == expected


@pytest.mark.parametrize("filtered", [False, True])
@pytest.mark.parametrize(
    "multi",
    [
        False,
        param(
            True,
            marks=[
                pytest.mark.notimpl(
                    ["datafusion", "oracle", "snowflake", "polars", "risingwave"],
                    raises=com.OperationNotDefinedError,
                    reason="multi-quantile not yet implemented",
                ),
                pytest.mark.notyet(
                    ["mssql", "exasol"],
                    raises=com.UnsupportedBackendType,
                    reason="array types not supported",
                ),
            ],
        ),
    ],
)
@pytest.mark.notyet(
    ["druid", "flink", "impala", "mysql", "sqlite"],
    raises=(com.OperationNotDefinedError, com.UnsupportedBackendType),
    reason="quantiles (approximate or otherwise) not supported",
)
def test_approx_quantile(con, filtered, multi):
    t = ibis.memtable({"x": [0, 25, 25, 50, 75, 75, 100, 125, 125, 150, 175, 175, 200]})
    where = t.x <= 100 if filtered else None
    q = [0.25, 0.75] if multi else 0.25
    res = con.execute(t.x.approx_quantile(q, where=where))
    if multi:
        assert isinstance(res, list)
        assert all(pd.api.types.is_float(r) for r in res)
        sol = [25, 75] if filtered else [50, 150]
    else:
        assert pd.api.types.is_float(res)
        sol = 25 if filtered else 50

    # Give pretty wide bounds for approximation - we're mostly testing that
    # the call is valid and the filtering logic is applied properly.
    assert res == pytest.approx(sol, abs=10)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t, where: t.G.cov(t.RBI, where=where, how="pop"),
            lambda t, where: t.G[where].cov(t.RBI[where], ddof=0),
            id="covar_pop",
            marks=[
                pytest.mark.notimpl(
                    ["polars", "druid"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["mysql", "impala", "sqlite", "flink"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function covar_pop(integer, integer) does not exist",
                ),
            ],
        ),
        param(
            lambda t, where: t.G.cov(t.RBI, where=where, how="sample"),
            lambda t, where: t.G[where].cov(t.RBI[where], ddof=1),
            id="covar_samp",
            marks=[
                pytest.mark.notimpl(
                    ["polars", "druid"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["mysql", "impala", "sqlite", "flink"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function covar_pop(integer, integer) does not exist",
                ),
            ],
        ),
        param(
            lambda t, where: t.G.corr(t.RBI, where=where, how="pop"),
            lambda t, where: t.G[where].corr(t.RBI[where]),
            id="corr_pop",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["impala", "mysql", "sqlite", "flink"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["clickhouse"],
                    raises=(ValueError, AttributeError),
                    reason="ClickHouse only implements `sample` correlation coefficient",
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function covar_pop(integer, integer) does not exist",
                ),
            ],
        ),
        param(
            lambda t, where: t.G.corr(t.RBI, where=where, how="sample"),
            lambda t, where: t.G[where].corr(t.RBI[where]),
            id="corr_samp",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["postgres", "duckdb", "snowflake", "risingwave", "exasol"],
                    raises=com.UnsupportedOperationError,
                    reason="backend only implements population correlation coefficient",
                ),
                pytest.mark.notyet(
                    ["impala", "mysql", "sqlite", "flink"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=ValueError,
                    reason="Correlation with how='sample' is not supported.",
                ),
                pytest.mark.notyet(
                    [
                        "risingwave",
                        "snowflake",
                        "oracle",
                    ],
                    raises=ValueError,
                    reason="XXXXSQLExprTranslator only implements population correlation coefficient",
                ),
                pytest.mark.notyet(["trino"], raises=com.UnsupportedOperationError),
            ],
        ),
        param(
            lambda t, where: (t.G > 34.0).cov(
                t.G <= 34.0,
                where=where,
                how="pop",
            ),
            lambda t, where: (t.G[where] > 34.0).cov(t.G[where] <= 34.0, ddof=0),
            id="covar_pop_bool",
            marks=[
                pytest.mark.notimpl(
                    ["polars", "druid"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["mysql", "impala", "sqlite", "flink"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function covar_pop(integer, integer) does not exist",
                ),
            ],
        ),
        param(
            lambda t, where: (t.G > 34.0).corr(
                t.G <= 34.0,
                where=where,
                how="pop",
            ),
            lambda t, where: (t.G[where] > 34.0).corr(t.G[where] <= 34.0),
            id="corr_pop_bool",
            marks=[
                pytest.mark.notimpl(
                    ["druid"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["impala", "mysql", "sqlite", "flink"],
                    raises=com.OperationNotDefinedError,
                ),
                pytest.mark.notyet(
                    ["clickhouse"],
                    raises=ValueError,
                    reason="ClickHouse only implements `sample` correlation coefficient",
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function covar_pop(integer, integer) does not exist",
                ),
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.yearID.isin([2009, 2015]),
            lambda t: t.yearID.isin([2009, 2015]),
            id="cond",
        ),
    ],
)
@pytest.mark.notimpl(["mssql"], raises=com.OperationNotDefinedError)
def test_corr_cov(
    con,
    batting,
    batting_df,
    result_fn,
    expected_fn,
    ibis_cond,
    pandas_cond,
):
    expr = result_fn(batting, ibis_cond(batting)).name("tmp")
    result = expr.execute()

    expected = expected_fn(batting_df, pandas_cond(batting_df))

    # Backends use different algorithms for computing covariance each with
    # different amounts of numerical stability.
    #
    # This makes a generic, precise and accurate comparison function incredibly
    # fragile and tedious to write.
    assert (
        # polars seems to have a numerically-imprecise-compared-to-pandas
        # (though perhaps fast!) implementation of correlation and covariance
        pytest.approx(result, abs=1e-2 if con.name == "polars" else None) == expected
    )


@pytest.mark.notimpl(
    ["mysql", "sqlite", "mssql", "druid"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notyet(["flink"], raises=com.OperationNotDefinedError)
def test_approx_median(alltypes):
    expr = alltypes.double_col.approx_median()
    result = expr.execute()
    assert isinstance(result, float)


@pytest.mark.notimpl(
    ["bigquery", "druid", "sqlite"], raises=com.OperationNotDefinedError
)
@pytest.mark.notyet(
    ["impala", "mysql", "mssql", "druid", "trino"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.never(
    ["flink"],
    reason="backend doesn't implement approximate quantiles yet",
    raises=com.OperationNotDefinedError,
)
def test_median(alltypes, df):
    expr = alltypes.double_col.median()
    result = expr.execute()
    expected = df.double_col.median()
    assert result == expected


@pytest.mark.notimpl(
    ["bigquery", "druid", "sqlite"], raises=com.OperationNotDefinedError
)
@pytest.mark.notyet(
    ["impala", "mysql", "mssql", "trino", "flink"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["clickhouse"],
    raises=ClickHouseDatabaseError,
    reason="doesn't support median of strings",
)
@pytest.mark.notyet(
    ["pyspark"], raises=AssertionError, reason="pyspark returns null for string median"
)
@pytest.mark.notyet(
    ["databricks"],
    raises=DatabricksServerOperationError,
    reason="percentile of string is not allowed",
)
@pytest.mark.notyet(
    ["snowflake"],
    raises=SnowflakeProgrammingError,
    reason="doesn't support median of strings",
)
@pytest.mark.notyet(
    ["exasol"],
    raises=ExaQueryError,
    reason="doesn't support quantile on strings",
)
@pytest.mark.notyet(["polars"], raises=PolarsInvalidOperationError)
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="not supported upstream")
@pytest.mark.parametrize(
    "func",
    [
        param(methodcaller("quantile", 0.5), id="quantile"),
        param(
            methodcaller("median"),
            id="median",
            marks=[
                pytest.mark.notyet(
                    ["oracle"],
                    raises=OracleDatabaseError,
                    reason="doesn't support median of strings",
                )
            ],
        ),
    ],
)
def test_string_quantile(alltypes, func):
    expr = func(alltypes.select(col=ibis.literal("a")).limit(5).col)
    result = expr.execute()
    assert result == "a"


@pytest.mark.notimpl(
    ["bigquery", "sqlite", "druid"], raises=com.OperationNotDefinedError
)
@pytest.mark.notyet(
    ["impala", "mysql", "mssql", "trino", "exasol", "flink"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["snowflake"],
    raises=SnowflakeProgrammingError,
    reason="doesn't support median of dates",
)
@pytest.mark.notyet(
    ["pyspark"],
    raises=PySparkAnalysisException,
    reason="doesn't support quantile on dates",
)
@pytest.mark.notyet(
    ["exasol"],
    raises=ExaQueryError,
    reason="doesn't support quantile on dates",
)
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="not supported upstream")
@pytest.mark.notyet(
    ["polars"], raises=PolarsInvalidOperationError, reason="not supported upstream"
)
@pytest.mark.notyet(
    ["databricks"],
    raises=DatabricksServerOperationError,
    reason="percentile of string is not allowed",
)
def test_date_quantile(alltypes):
    expr = alltypes.timestamp_col.date().quantile(0.5)
    result = expr.execute()
    assert result == date(2009, 12, 31)


@pytest.mark.parametrize(
    ("ibis_sep", "pandas_sep"),
    [
        param(":", ":", id="const"),
        param(
            L(":") + ":",
            "::",
            id="expr",
            marks=[
                pytest.mark.notyet(["mysql"], raises=com.UnsupportedOperationError),
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="Argument 2 to STRING_AGG must be a literal or query parameter",
                ),
                pytest.mark.notimpl(
                    ["polars"],
                    raises=com.UnsupportedArgumentError,
                    reason="polars doesn't support expression separators",
                ),
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    ("ibis_cond", "pandas_cond"),
    [
        param(lambda _: None, lambda _: slice(None), id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
        param(
            lambda t: t.string_col.notin(["1", "7"]),
            lambda t: ~t.string_col.isin(["1", "7"]),
            id="not_in",
        ),
    ],
)
@pytest.mark.notyet(["flink"], raises=Py4JJavaError)
def test_group_concat(
    backend, alltypes, df, ibis_cond, pandas_cond, ibis_sep, pandas_sep
):
    expr = (
        alltypes.group_by("bigint_col")
        .aggregate(
            tmp=lambda t: t.string_col.group_concat(ibis_sep, where=ibis_cond(t))
        )
        .order_by("bigint_col")
    )
    result = expr.execute()
    expected = (
        (
            df
            if isinstance(pandas_cond(df), slice)
            else df.assign(string_col=df.string_col.where(pandas_cond(df)))
        )
        .groupby("bigint_col")
        .string_col.agg(
            lambda s: (np.nan if pd.isna(s).all() else pandas_sep.join(s.values))
        )
        .rename("tmp")
        .sort_index()
        .reset_index()
    )

    backend.assert_frame_equal(
        result.replace(np.nan, None), expected.replace(np.nan, None)
    )


@pytest.mark.notimpl(
    [
        "clickhouse",
        "datafusion",
        "druid",
        "flink",
        "impala",
        "pyspark",
        "sqlite",
        "databricks",
    ],
    raises=com.UnsupportedOperationError,
)
@pytest.mark.parametrize("filtered", [False, True])
def test_group_concat_ordered(alltypes, df, filtered):
    ibis_cond = (_.id % 13 == 0) if filtered else None
    pd_cond = (df.id % 13 == 0) if filtered else True
    expr = (
        alltypes.filter(_.bigint_col == 10)
        .id.cast("str")
        .group_concat(":", where=ibis_cond, order_by=_.id.desc())
    )
    result = expr.execute()
    expected = ":".join(
        df.id[(df.bigint_col == 10) & pd_cond].sort_values(ascending=False).astype(str)
    )
    assert result == expected


def gen_test_collect_marks(distinct, filtered, ordered, include_null):
    """The marks for this test fail for different combinations of parameters.
    Rather than set `strict=False` (which can let bugs sneak through), we split
    the mark generation into a function"""
    if distinct:
        yield pytest.mark.notimpl(["datafusion"], raises=com.UnsupportedOperationError)
    if ordered:
        yield pytest.mark.notimpl(
            ["clickhouse", "pyspark", "flink", "databricks"],
            raises=com.UnsupportedOperationError,
        )
    if include_null:
        yield pytest.mark.notimpl(
            ["clickhouse", "pyspark", "snowflake", "databricks"],
            raises=com.UnsupportedOperationError,
        )

    # Handle special cases
    if filtered and distinct:
        yield pytest.mark.notimpl(
            ["bigquery", "snowflake"],
            raises=com.UnsupportedOperationError,
            reason="Can't combine where and distinct",
        )
    elif filtered and include_null:
        yield pytest.mark.notimpl(
            ["bigquery"],
            raises=com.UnsupportedOperationError,
            reason="Can't combine where and include_null",
        )
    elif include_null:
        yield pytest.mark.notimpl(
            ["bigquery"],
            raises=GoogleBadRequest,
            reason="BigQuery can't retrieve arrays with null values",
        )


@pytest.mark.notimpl(
    ["druid", "exasol", "impala", "mssql", "mysql", "oracle", "sqlite"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.parametrize(
    "distinct, filtered, ordered, include_null",
    [
        param(*ps, marks=list(gen_test_collect_marks(*ps)))
        for ps in itertools.product(*([[True, False]] * 4))
    ],
)
def test_collect(alltypes, df, distinct, filtered, ordered, include_null):
    expr = alltypes.mutate(x=_.string_col.nullif("3")).x.collect(
        where=((_.id % 13 == 0) if filtered else None),
        include_null=include_null,
        distinct=distinct,
        order_by=(_.x.desc() if ordered else ()),
    )
    res = expr.execute()

    x = df.string_col.where(df.string_col != "3", None)
    if filtered:
        x = x[df.id % 13 == 0]
    if not include_null:
        x = x.dropna()
    if distinct:
        x = x.drop_duplicates()
    sol = sorted(x, key=lambda x: (x is not None, x), reverse=True)

    if not ordered:
        # If unordered, order afterwards so we can compare
        res = sorted(res, key=lambda x: (x is not None, x), reverse=True)

    assert res == sol


@pytest.mark.notimpl(["mssql"], raises=PyODBCProgrammingError)
def test_topk_op(alltypes, df):
    # TopK expression will order rows by "count" but each backend
    # can have different result for that.
    # Note: Maybe would be good if TopK could order by "count"
    # and the field used by TopK
    t = alltypes.order_by(alltypes.string_col)
    df = df.sort_values("string_col")
    expr = t.string_col.topk(3)
    result = expr.execute()
    expected = df.groupby("string_col")["string_col"].count().head(3)
    assert all(result.iloc[:, 1].values == expected.values)


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        param(
            lambda t: t.semi_join(t.string_col.topk(3), "string_col"),
            lambda t: t[
                t.string_col.isin(
                    t.groupby("string_col")["string_col"].count().head(3).index
                )
            ],
            id="string_col_filter_top3",
            marks=pytest.mark.notimpl(["mssql"], raises=PyODBCProgrammingError),
        )
    ],
)
@pytest.mark.notyet(
    ["druid"], raises=PyDruidProgrammingError, reason="Java NullPointerException"
)
@pytest.mark.notyet(
    ["flink"], raises=Py4JError, reason="Flink doesn't support semi joins"
)
def test_topk_filter_op(con, alltypes, df, result_fn, expected_fn):
    # TopK expression will order rows by "count" but each backend
    # can have different result for that.
    # Note: Maybe would be good if TopK could order by "count"
    # and the field used by TopK
    if con.name == "sqlite":
        # TODO: remove after CTE extraction is reimplemented
        pytest.skip("topk -> semi-join performance has increased post SQLGlot refactor")
    t = alltypes.order_by(alltypes.string_col)
    df = df.sort_values("string_col")
    expr = result_fn(t)
    result = expr.execute()
    expected = expected_fn(df)
    assert result.shape[0] == expected.shape[0]


@pytest.mark.parametrize(
    "agg_fn", [lambda s: list(s), lambda s: np.array(s)], ids=["list", "ndarray"]
)
@pytest.mark.notimpl(
    [
        "bigquery",
        "clickhouse",
        "datafusion",
        "duckdb",
        "impala",
        "mysql",
        "postgres",
        "risingwave",
        "sqlite",
        "snowflake",
        "polars",
        "mssql",
        "trino",
        "druid",
        "oracle",
        "exasol",
        "flink",
        "databricks",
    ],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["pyspark"],
    condition=IS_SPARK_REMOTE,
    raises=PySparkPythonException,
    reason="remote udfs not yet tested due to environment complexities",
)
def test_aggregate_list_like(backend, alltypes, df, agg_fn):
    """Tests .aggregate() where the result of an aggregation is a list-like.

    We expect the list / np.array to be treated as a scalar (in other
    words, the resulting table expression should have one element, which
    is the list / np.array).
    """
    with pytest.warns(FutureWarning, match="v9.0"):
        udf = reduction(input_type=[dt.double], output_type=dt.Array(dt.double))(agg_fn)

    expr = alltypes.aggregate(result_col=udf(alltypes.double_col))
    result = expr.execute()

    # Expecting a 1-row DataFrame
    expected = pd.DataFrame({"result_col": [agg_fn(df.double_col)]})

    backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(
    [
        "bigquery",
        "clickhouse",
        "datafusion",
        "duckdb",
        "impala",
        "mysql",
        "postgres",
        "risingwave",
        "sqlite",
        "snowflake",
        "polars",
        "mssql",
        "trino",
        "druid",
        "oracle",
        "flink",
        "exasol",
        "flink",
        "databricks",
    ],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["pyspark"],
    condition=IS_SPARK_REMOTE,
    raises=PySparkPythonException,
    reason="remote udfs not yet tested due to environment complexities",
)
def test_aggregate_mixed_udf(backend, alltypes, df):
    """Tests .aggregate() with multiple aggregations with mixed result types.

    (In particular, one aggregation that results in an array, and other
    aggregation(s) that result in a non-array)
    """
    with pytest.warns(FutureWarning, match="v9.0"):

        @reduction(input_type=[dt.double], output_type=dt.double)
        def sum_udf(v):
            return np.sum(v)

        @reduction(input_type=[dt.double], output_type=dt.Array(dt.double))
        def collect_udf(v):
            return np.array(v)

    expr = alltypes.aggregate(
        sum_col=sum_udf(alltypes.double_col),
        collect_udf=collect_udf(alltypes.double_col),
    )
    result = expr.execute()

    expected = pd.DataFrame(
        {
            "sum_col": [sum_udf.func(df.double_col)],
            "collect_udf": [collect_udf.func(df.double_col)],
        }
    )

    backend.assert_frame_equal(result, expected, check_like=True)


def test_binds_are_cast(alltypes):
    expr = alltypes.aggregate(
        high_line_count=alltypes.string_col.cases(("1-URGENT", 1), else_=0).sum()
    )

    expr.execute()


def test_agg_sort(alltypes):
    query = alltypes.aggregate(count=alltypes.count())
    with pytest.raises(com.IntegrityError):
        query.order_by(alltypes.year)


def test_filter(backend, alltypes, df):
    expr = (
        alltypes.filter(_.string_col == "1")
        .mutate(x=L(1, "int64"))
        .group_by(_.x)
        .aggregate(sum=_.double_col.sum())
    )

    # TODO: The pyspark backend doesn't apply schemas to outputs
    result = expr.execute().astype({"x": "int64"})
    expected = (
        df.loc[df.string_col == "1", :]
        .assign(x=1)
        .groupby("x")
        .double_col.sum()
        .rename("sum")
        .reset_index()
    )
    backend.assert_frame_equal(result, expected, check_like=True)


def test_agg_name_in_output_column(alltypes):
    query = alltypes.aggregate([alltypes.int_col.min(), alltypes.int_col.max()])
    df = query.execute()
    assert "min" in df.columns[0].lower()
    assert "max" in df.columns[1].lower()


def test_grouped_case(backend, con):
    table = ibis.memtable({"key": [1, 1, 2, 2], "value": [10, 30, 20, 40]})

    case_expr = ibis.cases((table.value < 25, table.value), else_=ibis.null())

    expr = (
        table.group_by(k="key")
        .aggregate(mx=case_expr.max())
        .drop_null("k")
        .order_by("k")
    )
    result = con.execute(expr)
    expected = pd.DataFrame({"k": [1, 2], "mx": [10, 20]})
    backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["polars"], raises=com.OperationNotDefinedError)
@pytest.mark.notimpl(
    ["datafusion"],
    raises=BaseException,
    reason="because pyo3 panic exception is raised",
)
@pytest.mark.notyet(["flink"], raises=Py4JJavaError)
@pytest.mark.notyet(["impala"], raises=ImpalaHiveServer2Error)
@pytest.mark.notyet(["clickhouse"], raises=ClickHouseDatabaseError)
@pytest.mark.notyet(["druid"], raises=PyDruidProgrammingError)
@pytest.mark.notyet(["snowflake"], raises=SnowflakeProgrammingError)
@pytest.mark.notyet(["trino"], raises=TrinoUserError)
@pytest.mark.notyet(["mysql"], raises=MySQLNotSupportedError)
@pytest.mark.notyet(["oracle"], raises=OracleDatabaseError)
@pytest.mark.notyet(["pyspark"], raises=PySparkAnalysisException)
@pytest.mark.notyet(["mssql"], raises=PyODBCProgrammingError)
@pytest.mark.notyet(["risingwave"], raises=AssertionError, strict=False)
@pytest.mark.notyet(["databricks"], raises=DatabricksServerOperationError)
def test_group_concat_over_window(backend, con):
    # TODO: this test is flaky on risingwave and I DO NOT LIKE IT
    input_df = pd.DataFrame(
        {
            "s": ["a|b|c", "b|a|c", "b|b|b|c|a"],
            "token": ["a", "b", "c"],
            "pk": [1, 1, 2],
            "id": [1, 2, 3],
        }
    )
    expected = input_df.assign(test=["a|b|c|b|a|c", "b|a|c", "b|b|b|c|a"])

    table = ibis.memtable(input_df)
    expr = table.mutate(
        test=table.s.group_concat(sep="|").over(
            group_by="pk", order_by="id", rows=(0, None)
        )
    ).order_by("id")

    result = con.execute(expr)
    backend.assert_frame_equal(result, expected)


def test_value_counts_on_expr(backend, alltypes, df):
    expr = alltypes.bigint_col.add(1).value_counts()
    columns = list(expr.columns)
    expr = expr.order_by(columns)
    result = expr.execute().sort_values(columns).reset_index(drop=True)
    expected = df.bigint_col.add(1).value_counts().reset_index()
    expected.columns = columns
    expected = expected.sort_values(by=columns).reset_index(drop=True)
    backend.assert_frame_equal(result, expected)


@pytest.mark.xfail_version(datafusion=["datafusion==42"])
def test_group_by_expr(backend, con):
    expr = (
        ibis.memtable(
            {"project_name": ["duckdb", "ibis-framework", "ibis", "numpy", "pandas"]}
        )
        .group_by(n=_.project_name.length())
        .aggregate(c=_.project_name.nunique())
        .order_by(_.n.desc())
    )
    result = con.execute(expr)
    expected = pd.DataFrame(dict(n=[14, 6, 5, 4], c=[1, 2, 1, 1])).astype(
        dict(n="int32", c="int64")
    )
    backend.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "value", [ibis.literal("a"), ibis.null("str")], ids=["string", "null"]
)
@pytest.mark.notyet(
    ["mssql"], raises=PyODBCProgrammingError, reason="not supported by the database"
)
def test_group_by_scalar(alltypes, df, value):
    expr = alltypes.group_by(key=value).agg(n=lambda t: t.count())
    result = expr.execute()
    n = result["n"].values[0].item()
    assert n == len(df)



from __future__ import annotations

import contextlib
import datetime

from ibis.common.temporal import normalize_timezone
from ibis.formats.pandas import PandasData


class ClickHousePandasData(PandasData):
    @classmethod
    def convert_Timestamp_element(cls, dtype):
        def converter(value, dtype=dtype):
            if value is None:
                return value

            with contextlib.suppress(AttributeError):
                value = value.item()

            if isinstance(value, int):
                # this can only mean a numpy or pandas timestamp because they
                # both support nanosecond precision
                #
                # when the precision is less than or equal to the value
                # supported by Python datetime.dateimte a call to .item() will
                # return a datetime.datetime but when the precision is higher
                # than the value supported by Python the value is an integer
                #
                # TODO: can we do better than implicit truncation to microseconds?
                import dateutil

                value = datetime.datetime.fromtimestamp(value / 1e9, dateutil.tz.UTC)

            if (tz := dtype.timezone) is not None:
                ntz = normalize_timezone(tz)

                # deal with this madness
                # https://github.com/ClickHouse/clickhouse-connect/pull/311
                if tz == "UTC" and value.tzinfo is None:
                    return value.replace(tzinfo=ntz)
                else:
                    return value.astimezone(ntz)

            return value.replace(tzinfo=None)

        return converter



from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.sql.compilers import PostgresCompiler
from ibis.backends.sql.compilers.base import ALL_OPERATIONS, NULL
from ibis.backends.sql.datatypes import RisingWaveType
from ibis.backends.sql.dialects import RisingWave


class RisingWaveCompiler(PostgresCompiler):
    __slots__ = ()

    dialect = RisingWave
    type_mapper = RisingWaveType

    LOWERED_OPS = {ops.Sample: None}

    UNSUPPORTED_OPS = (
        ops.Arbitrary,
        ops.Mode,
        ops.RandomScalar,
        ops.RandomUUID,
        ops.MultiQuantile,
        ops.ApproxMultiQuantile,
        ops.Sample,
        *(
            op
            for op in ALL_OPERATIONS
            if issubclass(
                op, (ops.GeoSpatialUnOp, ops.GeoSpatialBinOp, ops.GeoUnaryUnion)
            )
        ),
    )

    def visit_DateNow(self, op):
        return self.cast(sge.CurrentTimestamp(), dt.date)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if not order_by:
            raise com.UnsupportedOperationError(
                "RisingWave requires an `order_by` be specified in `first`"
            )
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.first_value(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if not order_by:
            raise com.UnsupportedOperationError(
                "RisingWave requires an `order_by` be specified in `last`"
            )
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.last_value(arg, where=where, order_by=order_by)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                "RisingWave only implements `pop` correlation coefficient"
            )
        return super().visit_Correlation(
            op, left=left, right=right, how=how, where=where
        )

    def visit_Quantile(self, op, *, arg, quantile, where):
        if where is not None:
            arg = self.if_(where, arg, NULL)
        suffix = "cont" if op.arg.dtype.is_numeric() else "disc"
        return sge.WithinGroup(
            this=self.f[f"percentile_{suffix}"](quantile),
            expression=sge.Order(expressions=[sge.Ordered(this=arg)]),
        )

    visit_ApproxQuantile = visit_Quantile

    def visit_TimestampTruncate(self, op, *, arg, unit):
        unit_mapping = {
            "Y": "year",
            "Q": "quarter",
            "M": "month",
            "W": "week",
            "D": "day",
            "h": "hour",
            "m": "minute",
            "s": "second",
            "ms": "milliseconds",
            "us": "microseconds",
        }

        if (unit := unit_mapping.get(unit.short)) is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {unit}")

        return self.f.date_trunc(unit, arg)

    visit_TimeTruncate = visit_DateTruncate = visit_TimestampTruncate

    def _make_interval(self, arg, unit):
        return arg * sge.Interval(this=sge.convert(1), unit=self.v[unit.name])

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.cast("".join(map(r"\x{:0>2x}".format, value)), dt.binary)
        elif dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_json():
            return sge.convert(str(value))
        return None


compiler = RisingWaveCompiler()



from __future__ import annotations

import concurrent.futures
import getpass
import sys
from os import environ as env
from typing import TYPE_CHECKING, Any

import ibis
from ibis.backends.tests.base import BackendTest

if TYPE_CHECKING:
    from ibis.backends import BaseBackend


def put_into(con, query):
    with con.cursor() as cur:
        cur.execute(query)


class TestConf(BackendTest):
    supports_map = True
    driver_supports_multiple_statements = False
    deps = ("databricks.sql",)

    def _load_data(self, **_: Any) -> None:
        import databricks.sql

        files = list(self.data_dir.joinpath("parquet").glob("*.parquet"))

        user = getpass.getuser()
        python_version = "".join(map(str, sys.version_info[:3]))
        volume = f"{user}_{python_version}"
        volume_prefix = f"/Volumes/ibis_testing/default/{volume}"

        with databricks.sql.connect(
            server_hostname=env["DATABRICKS_SERVER_HOSTNAME"],
            http_path=env["DATABRICKS_HTTP_PATH"],
            access_token=env["DATABRICKS_TOKEN"],
            staging_allowed_local_path=str(self.data_dir),
        ) as con:
            with con.cursor() as cur:
                cur.execute(
                    f"CREATE VOLUME IF NOT EXISTS ibis_testing.default.{volume} COMMENT 'Ibis test data storage'"
                )
            with concurrent.futures.ThreadPoolExecutor() as exe:
                for fut in concurrent.futures.as_completed(
                    exe.submit(
                        put_into,
                        con,
                        f"PUT '{file}' INTO '{volume_prefix}/{file.name}' OVERWRITE",
                    )
                    for file in files
                ):
                    fut.result()

            with con.cursor() as cur:
                for raw_stmt in self.ddl_script:
                    try:
                        stmt = raw_stmt.format(user=user, python_version=python_version)
                    except KeyError:  # not a valid format string, just execute it
                        stmt = raw_stmt

                    cur.execute(stmt)

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw) -> BaseBackend:
        return ibis.databricks.connect(
            server_hostname=env["DATABRICKS_SERVER_HOSTNAME"],
            http_path=env["DATABRICKS_HTTP_PATH"],
            access_token=env["DATABRICKS_TOKEN"],
            catalog="ibis_testing",
            schema="default",
            **kw,
        )



# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import literal


def test_lower_upper(table):
    lresult = table.g.lower()
    uresult = table.g.upper()

    assert isinstance(lresult, ir.StringColumn)
    assert isinstance(uresult, ir.StringColumn)

    assert isinstance(lresult.op(), ops.Lowercase)
    assert isinstance(uresult.op(), ops.Uppercase)

    lit = literal("FoO")

    lresult = lit.lower()
    uresult = lit.upper()
    assert isinstance(lresult, ir.StringScalar)
    assert isinstance(uresult, ir.StringScalar)


def test_substr(table):
    lit = literal("FoO")

    result = table.g.substr(2, 4)
    lit_result = lit.substr(0, 2)

    assert isinstance(result, ir.StringColumn)
    assert isinstance(lit_result, ir.StringScalar)

    op = result.op()
    assert isinstance(op, ops.Substring)

    start, length = op.args[1:]

    assert start == literal(2).op()
    assert length == literal(4).op()


def test_left_right(table):
    result = table.g.left(5)
    expected = table.g.substr(0, 5)
    assert result.equals(expected)

    result = table.g.right(5)
    op = result.op()
    assert isinstance(op, ops.StrRight)
    assert op.args[1] == literal(5).op()


def test_length(table):
    lit = literal("FoO")
    result = table.g.length()
    lit_result = lit.length()

    assert isinstance(result, ir.IntegerColumn)
    assert isinstance(lit_result, ir.IntegerScalar)
    assert isinstance(result.op(), ops.StringLength)


def test_join(table):
    dash = literal("-")

    expr = dash.join([table.f.cast("string"), table.g])
    assert isinstance(expr, ir.StringColumn)

    expr = dash.join([literal("ab"), literal("cd")])
    assert isinstance(expr, ir.StringScalar)


def test_contains(table):
    expr = table.g.contains("foo")
    assert isinstance(expr.op(), ops.StringContains)

    with pytest.raises(TypeError):
        "foo" in table.g  # noqa: B015


def test_add_radd(table, string_col):
    string_col = table[string_col]
    assert isinstance(literal("foo") + "bar", ir.StringScalar)
    assert isinstance("bar" + literal("foo"), ir.StringScalar)
    assert isinstance(string_col + "bar", ir.StringColumn)
    assert isinstance("bar" + string_col, ir.StringColumn)


def test_startswith(table):
    assert isinstance(table.g.startswith("foo"), ir.BooleanColumn)
    assert isinstance(literal("bar").startswith("foo"), ir.BooleanScalar)


def test_endswith(table):
    assert isinstance(table.g.endswith("foo"), ir.BooleanColumn)
    assert isinstance(literal("bar").endswith("foo"), ir.BooleanScalar)



from __future__ import annotations

import pytest
from pyflink.util.exceptions import TableException
from pytest import param

import ibis
from ibis import _
from ibis.backends.tests.errors import Py4JJavaError


@pytest.mark.xfail(raises=TableException)
def test_window_requires_order_by(con):
    t = con.tables.functional_alltypes
    expr = t.mutate(t.double_col - t.double_col.mean())
    con.execute(expr)


@pytest.mark.xfail(raises=TableException)
def test_window_does_not_support_multiple_order_by(con):
    t = con.tables.functional_alltypes
    expr = t.double_col.sum().over(rows=(-1, 1), order_by=[t.timestamp_col, t.int_col])
    con.execute(expr)


@pytest.mark.parametrize(
    "window",
    [
        param(
            {"rows": (-1, 1)},
            id="bounded_rows_following",
            marks=[pytest.mark.xfail(raises=TableException)],
        ),
        param(
            {"rows": (-1, None)},
            id="unbounded_rows_following",
            marks=[pytest.mark.xfail(raises=TableException)],
        ),
        param(
            {"rows": (-500, 1)},
            id="casted_bounded_rows_following",
            marks=[pytest.mark.xfail(raises=TableException)],
        ),
        param(
            {"range": (-1000, 0)},
            id="int_range",
            marks=[pytest.mark.xfail(raises=Py4JJavaError)],
        ),
    ],
)
def test_window_invalid_start_end(con, window):
    t = con.tables.functional_alltypes
    expr = t.int_col.sum().over(**window, order_by=t.timestamp_col)
    con.execute(expr)


def test_tumble_window_by_grouped_agg(con):
    t = con.table("functional_alltypes_with_watermark")
    expr = (
        t.window_by(time_col=t.timestamp_col)
        .tumble(size=ibis.interval(seconds=30))
        .agg(by=["string_col"], avg=_.float_col.mean())
    )
    result = expr.to_pandas()
    assert list(result.columns) == ["window_start", "window_end", "string_col", "avg"]
    assert result.shape == (610, 4)


def test_tumble_window_by_ungrouped_agg(con):
    t = con.table("functional_alltypes_with_watermark")
    expr = (
        t.window_by(time_col=t.timestamp_col)
        .tumble(size=ibis.interval(seconds=30))
        .agg(avg=_.float_col.mean())
    )
    result = expr.to_pandas()
    assert list(result.columns) == ["window_start", "window_end", "avg"]
    assert result.shape == (610, 3)



from __future__ import annotations

import contextlib
from collections.abc import Mapping

import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis import util
from ibis.backends.tests.errors import (
    DatabricksServerOperationError,
    PolarsColumnNotFoundError,
    PsycoPg2InternalError,
    PsycoPg2SyntaxError,
    Py4JJavaError,
    PySparkAnalysisException,
)
from ibis.common.exceptions import IbisError

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")
tm = pytest.importorskip("pandas.testing")

pytestmark = [
    pytest.mark.never(["mysql", "sqlite", "mssql"], reason="No struct support"),
    pytest.mark.notyet(["impala"]),
    pytest.mark.notimpl(["druid", "oracle", "exasol"]),
]


@pytest.mark.parametrize(
    ("field", "expected"),
    [
        param(
            "a",
            [1.0, 2.0, 2.0, 3.0, 3.0, np.nan, np.nan],
            id="a",
            marks=pytest.mark.notimpl(["snowflake"]),
        ),
        param(
            "b", ["apple", "banana", "banana", "orange", "orange", None, None], id="b"
        ),
        param(
            "c",
            [2, 2, 3, 3, 4, np.nan, np.nan],
            id="c",
            marks=pytest.mark.notimpl(["snowflake"]),
        ),
    ],
)
def test_single_field(struct, field, expected):
    expr = struct.select(field=lambda t: t.abc[field]).order_by("field")
    result = expr.execute()
    tm.assert_series_equal(result.field, pd.Series(expected, name="field"))


def test_all_fields(struct, struct_df):
    result = struct.abc.execute()
    expected = struct_df.abc

    assert {
        row if not isinstance(row, Mapping) else tuple(row.items()) for row in result
    } == {
        row if not isinstance(row, Mapping) else tuple(row.items()) for row in expected
    }


_SIMPLE_DICT = dict(a=1, b="2", c=3.0)
_STRUCT_LITERAL = ibis.struct(
    _SIMPLE_DICT,
    type="struct<a: int64, b: string, c: float64>",
)
_NULL_STRUCT_LITERAL = ibis.null().cast("struct<a: int64, b: string, c: float64>")


@pytest.mark.notimpl(["postgres", "risingwave"])
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="unsupported syntax")
@pytest.mark.parametrize("field", ["a", "b", "c"])
def test_literal(backend, con, field):
    query = _STRUCT_LITERAL[field]
    dtype = query.type().to_pandas()
    result = pd.Series([con.execute(query)], dtype=dtype)
    result = result.replace({np.nan: None})
    expected = pd.Series([_SIMPLE_DICT[field]])
    backend.assert_series_equal(result, expected.astype(dtype))


@pytest.mark.notimpl(["postgres"])
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="unsupported syntax")
@pytest.mark.parametrize("field", ["a", "b", "c"])
@pytest.mark.notyet(
    ["clickhouse"], reason="clickhouse doesn't support nullable nested types"
)
def test_null_literal(backend, con, field):
    query = _NULL_STRUCT_LITERAL[field]
    result = pd.Series([con.execute(query)])
    result = result.replace({np.nan: None})
    expected = pd.Series([None], dtype="object")
    backend.assert_series_equal(result, expected)


@pytest.mark.notimpl(["postgres", "risingwave"])
def test_struct_column(alltypes, df):
    t = alltypes
    expr = t.select(s=ibis.struct(dict(a=t.string_col, b=1, c=t.bigint_col)))
    assert expr.s.type() == dt.Struct(dict(a=dt.string, b=dt.int8, c=dt.int64))
    result = expr.execute()
    expected = pd.DataFrame(
        {"s": [dict(a=a, b=1, c=c) for a, c in zip(df.string_col, df.bigint_col)]}
    )
    tm.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["postgres", "risingwave", "polars"])
def test_collect_into_struct(alltypes):
    from ibis import _

    t = alltypes
    expr = (
        t.filter(_.string_col.isin(("0", "1")))
        .group_by(group="string_col")
        .agg(
            val=lambda t: ibis.struct(
                dict(key=t.bigint_col.collect().cast("!array<int64>"))
            )
        )
    )
    result = expr.execute()
    assert result.shape == (2, 2)
    assert set(result.group) == {"0", "1"}
    val = result.val
    assert len(val.loc[result.group == "0"].iat[0]["key"]) == 730
    assert len(val.loc[result.group == "1"].iat[0]["key"]) == 730


@pytest.mark.notimpl(
    ["postgres"], reason="struct literals not implemented", raises=PsycoPg2SyntaxError
)
@pytest.mark.notimpl(
    ["risingwave"],
    reason="struct literals not implemented",
    raises=PsycoPg2InternalError,
)
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="unsupported syntax")
@pytest.mark.notimpl(["flink"], raises=Py4JJavaError, reason="not implemented in ibis")
def test_field_access_after_case(con):
    s = ibis.struct({"a": 3})
    x = ibis.cases((True, s), else_=ibis.struct({"a": 4}))
    y = x.a
    assert con.to_pandas(y) == 3


@pytest.mark.notimpl(
    ["postgres"], reason="struct literals not implemented", raises=PsycoPg2SyntaxError
)
@pytest.mark.notimpl(["flink"], raises=IbisError, reason="not implemented in ibis")
@pytest.mark.parametrize(
    "nullable",
    [
        param(
            True,
            marks=pytest.mark.notyet(
                ["clickhouse"],
                raises=AssertionError,
                reason="clickhouse doesn't allow nullable nested types",
            ),
            id="nullable",
        ),
        param(
            False,
            marks=[
                pytest.mark.notyet(
                    ["polars"],
                    raises=AssertionError,
                    reason="polars doesn't support non-nullable types",
                ),
                pytest.mark.notyet(
                    ["risingwave"],
                    reason="non-nullable struct types not implemented",
                    raises=PsycoPg2InternalError,
                ),
                pytest.mark.notimpl(
                    ["pyspark"],
                    raises=AssertionError,
                    reason="non-nullable struct types not yet implemented in Ibis's PySpark backend",
                ),
            ],
            id="non-nullable",
        ),
    ],
)
@pytest.mark.notyet(
    ["trino"],
    raises=AssertionError,
    reason="trino returns unquoted and therefore unparsable struct field names, we fall back to dt.unknown",
)
@pytest.mark.notyet(
    ["snowflake"],
    raises=AssertionError,
    reason="snowflake doesn't have strongly typed structs",
)
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="unsupported syntax")
@pytest.mark.notyet(
    ["databricks"],
    raises=DatabricksServerOperationError,
    reason="spaces are not allowed in column names",
)
def test_keyword_fields(con, nullable):
    schema = ibis.schema(
        {
            "struct": dt.Struct(
                {
                    "select": "int",
                    "from": "int",
                    "where": "int",
                    "order": "int",
                    "left join": "int",
                    "full outer join": "int",
                },
                nullable=nullable,
            )
        }
    )

    name = util.gen_name("struct_keywords")
    t = con.create_table(name, schema=schema)

    try:
        assert t.schema() == schema
        assert t.count().execute() == 0
    finally:
        with contextlib.suppress(NotImplementedError):
            con.drop_table(name, force=True)


@pytest.mark.notyet(
    ["postgres"],
    raises=PsycoPg2SyntaxError,
    reason="sqlglot doesn't implement structs for postgres correctly",
)
@pytest.mark.notyet(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="sqlglot doesn't implement structs for postgres correctly",
)
@pytest.mark.notyet(
    ["polars"],
    raises=PolarsColumnNotFoundError,
    reason="doesn't seem to support IN-style subqueries on structs",
)
@pytest.mark.xfail_version(
    pyspark=["pyspark<3.5"],
    reason="requires pyspark 3.5",
    raises=PySparkAnalysisException,
)
@pytest.mark.notimpl(
    ["flink"],
    raises=Py4JJavaError,
    reason="fails to parse due to an unsupported operation; flink docs say the syntax is supported",
)
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="unsupported syntax")
def test_isin_struct(con):
    needle1 = ibis.struct({"x": 1, "y": 2})
    needle2 = ibis.struct({"x": 2, "y": 3})
    haystack_t = ibis.memtable({"xs": [1, 2, 3], "ys": [2, 3, 4]})
    haystack = ibis.struct({"x": haystack_t.xs, "y": haystack_t.ys})
    both = needle1.isin(haystack) | needle2.isin(haystack)
    result = con.execute(both)
    # TODO(cpcloud): ensure the type is consistent
    assert result is True or result is np.bool_(True)



from __future__ import annotations

from operator import methodcaller
from time import sleep

import pandas as pd
import pytest

from ibis.backends.pyspark.datatypes import PySparkSchema


@pytest.mark.parametrize(
    "method",
    [
        methodcaller("read_delta", path="test.delta"),
        methodcaller("read_csv", source_list="test.csv"),
        methodcaller("read_parquet", path="test.parquet"),
        methodcaller("read_json", source_list="test.json"),
    ],
)
def test_streaming_import_not_implemented(con_streaming, method):
    with pytest.raises(NotImplementedError):
        method(con_streaming)


def test_read_csv_dir(con_streaming, awards_players_schema):
    t = con_streaming.read_csv_dir(
        "ci/ibis-testing-data/directory/csv/awards_players",
        table_name="t",
        schema=PySparkSchema.from_ibis(awards_players_schema),
        header=True,
    )
    con_streaming.write_to_memory(t, "n")
    sleep(2)  # wait for results to populate; count(*) returns 0 if executed right away
    pd_df = con_streaming._session.sql("SELECT count(*) FROM n").toPandas()
    assert not pd_df.empty
    assert pd_df.iloc[0, 0] == 6078


def test_read_parquet_dir(con_streaming):
    t = con_streaming.read_parquet_dir(
        "ci/ibis-testing-data/directory/parquet/awards_players", table_name="t"
    )
    con_streaming.write_to_memory(t, "n")
    sleep(2)  # wait for results to populate; count(*) returns 0 if executed right away
    pd_df = con_streaming._session.sql("SELECT count(*) FROM n").toPandas()
    assert not pd_df.empty
    assert pd_df.iloc[0, 0] == 6078


def test_to_csv_dir(con_streaming, tmp_path):
    t = con_streaming.table("awards_players")
    path = tmp_path / "out"
    con_streaming.to_csv_dir(
        t.limit(5),
        path=path,
        options={"checkpointLocation": tmp_path / "checkpoint", "header": True},
    )
    sleep(2)
    df = pd.concat([pd.read_csv(f) for f in path.glob("*.csv")])
    assert len(df) == 5


def test_to_parquet_dir(con_streaming, tmp_path):
    t = con_streaming.table("awards_players")
    path = tmp_path / "out"
    con_streaming.to_parquet_dir(
        t.limit(5),
        path=path,
        options={"checkpointLocation": tmp_path / "checkpoint"},
    )
    sleep(2)
    df = pd.concat([pd.read_parquet(f) for f in path.glob("*.parquet")])
    assert len(df) == 5



from __future__ import annotations

import inspect

import pytest

from ibis.backends import (
    BaseBackend,
    CanCreateCatalog,
    CanCreateDatabase,
    CanListCatalog,
    CanListDatabase,
)
from ibis.backends.sql import SQLBackend
from ibis.backends.tests.signature.typecheck import compatible

SKIP_METHODS = ["do_connect", "from_connection"]


def _scrape_methods(modules, params):
    params = []
    for module in modules:
        methods = list(filter(lambda x: not x.startswith("_"), dir(module)))
        for method in methods:
            # Only test methods that are callable (so we can grab the signature)
            # and skip any methods that we don't want to check
            if (
                method not in SKIP_METHODS
                and method not in marks.keys()
                and callable(getattr(module, method))
            ):
                params.append((module, method))

    return params


marks = {
    "compile": pytest.param(
        BaseBackend,
        "compile",
        marks=pytest.mark.notyet(
            [
                "bigquery",
                "clickhouse",
                "datafusion",
                "druid",
                "duckdb",
                "exasol",
                "impala",
                "mssql",
                "mysql",
                "oracle",
                "postgres",
                "pyspark",
                "risingwave",
                "snowflake",
                "sqlite",
                "trino",
                "databricks",
            ],
            reason="SQL backends all have an additional `pretty` argument for formatting the generated SQL",
        ),
    ),
    "create_database": pytest.param(
        CanCreateDatabase,
        "create_database",
        marks=pytest.mark.notyet(["clickhouse", "flink", "impala", "mysql", "pyspark"]),
    ),
    "drop_database": pytest.param(
        CanCreateDatabase,
        "drop_database",
        marks=pytest.mark.notyet(["clickhouse", "impala", "mysql", "pyspark"]),
    ),
    "drop_table": pytest.param(
        SQLBackend,
        "drop_table",
        marks=pytest.mark.notyet(["bigquery", "druid", "flink", "impala", "polars"]),
    ),
    "execute": pytest.param(
        SQLBackend,
        "execute",
        marks=pytest.mark.notyet(["clickhouse", "datafusion", "flink", "mysql"]),
    ),
    "insert": pytest.param(
        SQLBackend,
        "insert",
        marks=pytest.mark.notyet(["clickhouse", "flink", "impala"]),
    ),
    "list_databases": pytest.param(
        CanCreateDatabase,
        "list_databases",
        marks=pytest.mark.notyet(
            [
                "clickhouse",
                "flink",
                "impala",
                "mysql",
                "postgres",
                "risingwave",
                "sqlite",
            ]
        ),
    ),
    "list_tables": pytest.param(
        BaseBackend,
        "list_tables",
        marks=pytest.mark.notyet(["flink"]),
    ),
    "read_csv": pytest.param(
        BaseBackend,
        "read_csv",
        marks=pytest.mark.notyet(["duckdb", "flink", "pyspark", "datafusion"]),
    ),
    "read_delta": pytest.param(
        BaseBackend,
        "read_delta",
        marks=pytest.mark.notyet(["datafusion", "duckdb", "polars", "pyspark"]),
    ),
    "read_json": pytest.param(
        BaseBackend,
        "read_json",
        marks=pytest.mark.notyet(["duckdb", "flink", "pyspark"]),
    ),
    "read_parquet": pytest.param(
        BaseBackend,
        "read_parquet",
        marks=pytest.mark.notyet(["duckdb", "flink"]),
    ),
    "to_parquet_dir": pytest.param(
        BaseBackend,
        "to_parquet_dir",
        marks=pytest.mark.notyet(["pyspark"]),
    ),
}

params = _scrape_methods(
    [
        BaseBackend,
        SQLBackend,
        CanCreateCatalog,
        CanCreateDatabase,
        CanListCatalog,
        CanListDatabase,
    ],
    marks,
)

params.extend(marks.values())


@pytest.mark.parametrize("base_cls, method", params)
def test_signatures(base_cls, method, backend_cls):
    if not hasattr(backend_cls, method):
        pytest.skip(f"Method {method} not present in {backend_cls}, skipping...")

    base_sig = inspect.signature(getattr(base_cls, method))
    backend_sig = inspect.signature(getattr(backend_cls, method))

    # Usage is compatible(implementation_signature, defined_interface_signature, ...)
    assert compatible(backend_sig, base_sig, check_annotations=False)



from __future__ import annotations

import os
import sqlite3

import duckdb
import numpy as np
import pandas as pd
import pandas.testing as tm
import pyarrow as pa
import pytest
from packaging.version import parse as parse_version
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis.conftest import ARM64, LINUX, MACOS, SANDBOXED


def test_read_csv(con, data_dir):
    t = con.read_csv(data_dir / "csv" / "functional_alltypes.csv")
    assert t.count().execute()


def test_read_csv_with_columns(con, data_dir):
    t = con.read_csv(
        data_dir / "csv" / "awards_players.csv",
        header=True,
        columns={
            "playerID": "VARCHAR",
            "awardID": "VARCHAR",
            "yearID": "DATE",
            "lgID": "VARCHAR",
            "tie": "VARCHAR",
            "notes": "VARCHAR",
        },
        dateformat="%Y",
    )
    assert t.count().execute()


def test_read_parquet(con, data_dir):
    t = con.read_parquet(data_dir / "parquet" / "functional_alltypes.parquet")
    assert t.count().execute()


@pytest.mark.xfail(
    LINUX and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
)
def test_load_spatial_when_geo_column(tmpdir):
    pytest.importorskip("geopandas")
    pytest.importorskip("shapely")

    path = str(tmpdir.join("test_load_spatial.ddb"))

    with duckdb.connect(
        # windows is horrible and cannot download in parallel without
        # clobbering existing files, so give a temporary custom directory for
        # extensions
        path,
        config={"extension_directory": str(tmpdir.join("extensions"))},
    ) as con:
        con.install_extension("spatial")
        con.load_extension("spatial")
        con.execute(
            # create a table with a geom column
            """
            CREATE or REPLACE TABLE samples (name VARCHAR, geom GEOMETRY);

            INSERT INTO samples VALUES
              ('Point', ST_GeomFromText('POINT(-100 40)')),
              ('Linestring', ST_GeomFromText('LINESTRING(0 0, 1 1, 2 1, 2 2)')),
              ('Polygon', ST_GeomFromText('POLYGON((0 0, 1 0, 1 1, 0 1, 0 0))'));
            """
        )

    # load data from ibis and check for spatial extension
    con = ibis.duckdb.connect(path)

    query = """\
SELECT extension_name AS name
FROM duckdb_extensions()
WHERE installed AND loaded"""

    assert "spatial" not in con.sql(query).name.to_pandas().values

    # trigger spatial extension load
    assert not con.tables.samples.head(1).geom.to_pandas().empty

    assert "spatial" in con.sql(query).name.to_pandas().values


@pytest.mark.usefixtures("gpd")
def test_read_geo_to_pyarrow(con, data_dir):
    shapely = pytest.importorskip("shapely")

    t = con.read_geo(data_dir / "geojson" / "zones.geojson")
    raw_geometry = t.head().to_pyarrow()["geom"].to_pandas()
    assert len(shapely.from_wkb(raw_geometry))


def test_read_geo_to_geopandas(con, data_dir, gpd):
    t = con.read_geo(data_dir / "geojson" / "zones.geojson")
    gdf = t.head().to_pandas()
    assert isinstance(gdf, gpd.GeoDataFrame)


def test_read_geo_from_url(monkeypatch):
    con = ibis.duckdb.connect()

    loaded_exts = []
    monkeypatch.setattr(con, "_load_extensions", lambda x, **_: loaded_exts.extend(x))

    with pytest.raises((duckdb.IOException, duckdb.CatalogException)):
        # The read will fail, either because the URL is bogus (which it is) or
        # because the current connection doesn't have the spatial extension
        # installed and so the call to `st_read` will raise a catalog error.
        con.read_geo("https://...")

    assert "spatial" in loaded_exts
    assert "httpfs" in loaded_exts


def test_read_json(con, data_dir, tmp_path):
    pqt = con.read_parquet(data_dir / "parquet" / "functional_alltypes.parquet")

    path = tmp_path.joinpath("ft.json")
    path.write_text(pqt.execute().to_json(orient="records", lines=True))

    jst = con.read_json(path)

    nrows = pqt.count().execute()
    assert nrows
    assert nrows == jst.count().execute()


@pytest.fixture(scope="session")
def pgurl():  # pragma: no cover
    pgcon = ibis.postgres.connect(
        user="postgres",
        password="postgres",  # noqa: S106
        host="localhost",
    )

    df = pd.DataFrame({"x": [1.0, 2.0, 3.0, 1.0], "y": ["a", "b", "c", "a"]})

    pgcon.create_table("duckdb_test", df, overwrite=True)
    yield pgcon.con.info

    pgcon.drop_table("duckdb_test", force=True)


@pytest.mark.skipif(
    os.environ.get("DUCKDB_POSTGRES") is None, reason="avoiding CI shenanigans"
)
def test_read_postgres(con, pgurl):  # pragma: no cover
    # we don't run this test in CI, only locally, to avoid bringing a postgres
    # container up just for this test.  To run locally set env variable to True
    # and once a postgres container is up run the test.
    table = con.read_postgres(
        f"postgres://{pgurl.user}:{pgurl.password}@{pgurl.host}:{pgurl.port}",
        table_name="duckdb_test",
    )
    assert table.count().execute()


@pytest.fixture(scope="session")
def mysqlurl():  # pragma: no cover
    mysqlcon = ibis.mysql.connect(
        user="ibis",
        password="ibis",  # noqa: S106
        database="ibis_testing",
    )

    df = pd.DataFrame({"x": [1.0, 2.0, 3.0, 1.0], "y": ["a", "b", "c", "a"]})
    s = ibis.schema(dict(x="float64", y="str"))

    mysqlcon.create_table("duckdb_test", df, schema=s, overwrite=True)
    yield mysqlcon.con
    mysqlcon.drop_table("duckdb_test", force=True)


@pytest.mark.skipif(
    os.environ.get("DUCKDB_MYSQL") is None, reason="avoiding CI shenanigans"
)
def test_read_mysql(con, mysqlurl):  # pragma: no cover
    # to run this test run first the mysql test suit to get the ibis-testing
    # we don't run this test in CI, only locally, to avoid bringing a mysql
    # container up just for this test.  To run locally set env variable to True
    # and once a mysql container is up run the test.

    # TODO(ncclementi) replace for mysqlurl.host when this is fix
    # https://github.com/duckdb/duckdb_mysql/issues/44
    hostname = "127.0.0.1"

    table = con.read_mysql(
        f"mysql://{mysqlurl.user.decode()}:{mysqlurl.password.decode()}@{hostname}:{mysqlurl.port}/ibis_testing",
        catalog="mysqldb",
        table_name="duckdb_test",
    )
    assert table.count().execute()


@pytest.mark.xfail(
    LINUX and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
    raises=duckdb.IOException,
)
def test_read_sqlite(con, tmp_path):
    path = tmp_path / "test.db"

    scon = sqlite3.connect(str(path))
    try:
        with scon:
            scon.execute("CREATE TABLE t AS SELECT 1 a UNION SELECT 2 UNION SELECT 3")
    finally:
        scon.close()

    ft = con.read_sqlite(path, table_name="t")
    assert ft.count().execute()


def test_read_sqlite_no_table_name(con, tmp_path):
    path = tmp_path / "test.db"

    scon = sqlite3.connect(str(path))
    try:
        assert path.exists()

        with pytest.raises(ValueError):
            con.read_sqlite(path)
    finally:
        scon.close()


# Because we create a new connection and the test requires loading/installing a
# DuckDB extension, we need to xfail these on Nix.
@pytest.mark.xfail(
    LINUX and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
    raises=duckdb.IOException,
)
def test_attach_sqlite(data_dir, tmp_path):
    import sqlite3

    # Create a new connection here because we already have the `ibis_testing`
    # tables loaded in to the `con` fixture.
    con = ibis.duckdb.connect()

    test_db_path = tmp_path / "test.db"
    scon = sqlite3.connect(test_db_path)
    try:
        with scon:
            scon.executescript((data_dir.parent / "schema" / "sqlite.sql").read_text())

        con.attach_sqlite(test_db_path)
        assert set(con.list_tables()) >= {
            "functional_alltypes",
            "awards_players",
            "batting",
            "diamonds",
        }

        fa = con.tables.functional_alltypes
        assert len(set(fa.schema().types)) > 1

        # overwrite existing sqlite_db and force schema to all strings
        con.attach_sqlite(test_db_path, overwrite=True, all_varchar=True)
        assert set(con.list_tables()) >= {
            "functional_alltypes",
            "awards_players",
            "batting",
            "diamonds",
        }

        fa = con.tables.functional_alltypes
        types = fa.schema().types
        assert len(set(types)) == 1
        assert dt.String(nullable=True) in set(types)
    finally:
        scon.close()


def test_memtable_with_nullable_dtypes(con):
    data = pd.DataFrame(
        {
            "a": pd.Series(["a", None, "c"], dtype="string"),
            "b": pd.Series([None, 1, 2], dtype="Int8"),
            "c": pd.Series([0, None, 2], dtype="Int16"),
            "d": pd.Series([0, 1, None], dtype="Int32"),
            "e": pd.Series([None, None, -1], dtype="Int64"),
            "f": pd.Series([None, 1, 2], dtype="UInt8"),
            "g": pd.Series([0, None, 2], dtype="UInt16"),
            "h": pd.Series([0, 1, None], dtype="UInt32"),
            "i": pd.Series([None, None, 42], dtype="UInt64"),
            "j": pd.Series([None, False, True], dtype="boolean"),
        }
    )
    expr = ibis.memtable(data)
    res = con.execute(expr)
    assert len(res) == len(data)


def test_memtable_with_nullable_pyarrow_string(con):
    pytest.importorskip("pyarrow")
    data = pd.DataFrame({"a": pd.Series(["a", None, "c"], dtype="string[pyarrow]")})
    expr = ibis.memtable(data)
    res = con.execute(expr)
    assert len(res) == len(data)


def test_memtable_with_nullable_pyarrow_not_string(con):
    pytest.importorskip("pyarrow")

    data = pd.DataFrame(
        {
            "b": pd.Series([None, 1, 2], dtype="int8[pyarrow]"),
            "c": pd.Series([0, None, 2], dtype="int16[pyarrow]"),
            "d": pd.Series([0, 1, None], dtype="int32[pyarrow]"),
            "e": pd.Series([None, None, -1], dtype="int64[pyarrow]"),
            "f": pd.Series([None, 1, 2], dtype="uint8[pyarrow]"),
            "g": pd.Series([0, None, 2], dtype="uint16[pyarrow]"),
            "h": pd.Series([0, 1, None], dtype="uint32[pyarrow]"),
            "i": pd.Series([None, None, 42], dtype="uint64[pyarrow]"),
            "j": pd.Series([None, False, True], dtype="boolean[pyarrow]"),
        }
    )
    expr = ibis.memtable(data)
    res = con.execute(expr)
    assert len(res) == len(data)


@pytest.mark.parametrize(
    "database",
    [lambda parent: parent / "test.ddb", lambda _: ":memory:"],
    ids=["disk", "memory"],
)
def test_temp_dir_set(tmp_path, database):
    temp_directory = tmp_path / "does" / "not" / "exist"
    temp_directory.mkdir(parents=True, exist_ok=True)
    con = ibis.duckdb.connect(database(tmp_path), temp_directory=temp_directory)
    assert con.settings["temp_directory"] == str(temp_directory)


@pytest.mark.xfail(
    SANDBOXED and LINUX,
    reason=(
        "nix on linux cannot download duckdb extensions or data due to sandboxing; "
        "duckdb will try to automatically install and load read_parquet"
    ),
    raises=(duckdb.Error, duckdb.IOException),
)
@pytest.mark.skipif(
    SANDBOXED and MACOS and ARM64, reason="raises a RuntimeError on nix macos arm64"
)
def test_s3_403_fallback(con, httpserver, monkeypatch):
    # monkeypatch to avoid downloading extensions in tests
    monkeypatch.setattr(con, "_load_extensions", lambda _: True)

    # Throw a 403 to trigger fallback to pyarrow.dataset
    path = "/invalid.parquet"
    httpserver.expect_request(path).respond_with_data(
        status=403, content_type="application/vnd.apache.parquet"
    )

    # Since the URI is nonsense to pyarrow, expect an error, but raises from
    # pyarrow, which indicates the fallback worked
    url = httpserver.url_for(path)
    with pytest.raises(pa.lib.ArrowInvalid):
        con.read_parquet(url)


def test_register_numpy_str(con):
    data = pd.DataFrame({"a": [np.str_("xyz"), None]})
    result = ibis.memtable(data)
    tm.assert_frame_equal(con.execute(result), data)


def test_memtable_recordbatchreader_raises(con):
    table = pa.Table.from_batches(
        map(pa.RecordBatch.from_pydict, [{"x": [1, 2]}, {"x": [3, 4]}])
    )
    reader = table.to_reader()

    with pytest.raises(TypeError):
        ibis.memtable(reader)

    t = ibis.memtable(reader.read_all())

    # First execute is fine
    res = con.execute(t)
    tm.assert_frame_equal(res, table.to_pandas())


def test_csv_with_slash_n_null(con, tmp_path):
    data_path = tmp_path / "data.csv"
    data_path.write_text("a\n1\n3\n\\N\n")
    t = con.read_csv(data_path, nullstr="\\N")
    col = t.a.execute()
    assert pd.isna(col.iat[-1])


@pytest.mark.xfail(
    LINUX and SANDBOXED, reason="nix can't hit GCS because it is sandboxed."
)
@pytest.mark.parametrize(
    "extensions",
    [
        param([], id="none"),
        param(
            ["httpfs"],
            marks=[
                pytest.mark.xfail(
                    parse_version("0.10.0")
                    <= parse_version(duckdb.__version__)
                    < parse_version("0.10.2"),
                    reason="https://github.com/duckdb/duckdb/issues/10698",
                    raises=duckdb.HTTPException,
                )
            ],
            id="httpfs",
        ),
    ],
)
def test_register_filesystem_gcs(extensions):
    fsspec = pytest.importorskip("fsspec")
    pytest.importorskip("gcsfs")

    con = ibis.duckdb.connect()

    for ext in extensions:
        con.load_extension(ext)

    gcs = fsspec.filesystem("gcs")

    con.register_filesystem(gcs)
    band_members = con.read_csv(
        "gcs://ibis-examples/data/band_members.csv.gz", table_name="band_members"
    )

    assert band_members.count().to_pyarrow()


def test_memtable_null_column_parquet_dtype_roundtrip(con, tmp_path):
    before = ibis.memtable({"a": [None, None, None]}, schema={"a": "string"})
    con.to_parquet(before, tmp_path / "tmp.parquet")
    after = con.read_parquet(tmp_path / "tmp.parquet")

    assert before.a.type() == after.a.type()


def test_read_json_no_auto_detection(con, tmp_path):
    ndjson_data = """
    {"year": 2007}
    {"year": 2008}
    {"year": 2009}
    """
    path = tmp_path.joinpath("test.ndjson")
    path.write_text(ndjson_data)

    t = con.read_json(path, auto_detect=False, columns={"year": "varchar"})
    assert t.year.type() == dt.string
